import json
import os
import re
import subprocess
import sys
import tarfile
import tempfile
import uuid
from pathlib import Path
from posixpath import basename
from signal import SIGKILL
from signal import Signals
from signal import valid_signals
from typing import Deque
from typing import Final
from typing import Iterable
from typing import List
from typing import Mapping
from typing import Optional
from typing import Tuple

from avalon.common.log_utils import log_to_sentry
from avalon.common.log_utils import logger
from avalon.common.utils import AVALON_PACKAGE_DIR
from avalon.common.utils import wait_until_true
from avalon.contrib.utils import FILESYSTEM_ROOT
from avalon.contrib.utils import TEMP_DIR
from avalon.datagen.data_config import AbstractDataConfig
from avalon.datagen.errors import GodotError
from avalon.datagen.godot_generated_types import READY_LOG_SIGNAL
from avalon.datagen.godot_generated_types import SimSpec
from avalon.datagen.world_creation.world_generator import GenerateAvalonWorldParams

# TODO: Consider linking binaries to `godot_editor` and `godot_runner` allowing for more fine-grained control/inspection from python
GODOT_BINARY_PATH_ENV_FLAG: Final = "GODOT_BINARY_PATH"
GODOT_BINARY_PATH: Final = os.environ.get(GODOT_BINARY_PATH_ENV_FLAG, f"{AVALON_PACKAGE_DIR}/bin/godot")

GODOT_EDITOR_PATH_ENV_FLAG: Final = "GODOT_EDITOR_PATH"
GODOT_EDITOR_PATH: Final = os.environ.get("GODOT_EDITOR_PATH", f"{GODOT_BINARY_PATH}_editor")

GODOT_ERROR_LOG_PATH: Final = f"{FILESYSTEM_ROOT}/godot"

AVALON_GODOT_PROJECT_PATH: Final = f"{AVALON_PACKAGE_DIR}/datagen/godot"
AVALON_HUMAN_WORLDS_PATH: Final = f"{AVALON_GODOT_PROJECT_PATH}/worlds"

DATAGEN_SCRIPT_PATH: Final = f"{AVALON_GODOT_PROJECT_PATH}/datagen.sh"

_ACTION_REPLAY_FILENAME: Final = "actions_replay.out"


def _create_json_from_config(config: AbstractDataConfig) -> str:
    return json.dumps(config.to_dict(), indent=2)


ERROR_ALLOWLIST = (
    "Basis must be normalized",
    "tree_exiting",
    "libgodot_openxr.so",
    "No library set for this platform",
)

SYSTEM_FINALIZING_NO_GODOT_RERAISE = "Could not check Godot log during Python finalization"


def _read_log(log_path: str) -> Tuple[List[str], bool]:
    """Read the godot log and check for errors"""
    try:
        with open(log_path, "r") as infile:
            log_lines = infile.readlines()
    except NameError:
        # We check Godot logs during GodotEnv.__del__ and depending on the exact script + Python internals, it may get
        # called during Python finalization, when open() and other builtins are not available. Can't use a custom
        # error class since they can't be caught during finalization.
        # This tweak can possibly be removed when we're running Python 3.10; ref: https://bugs.python.org/issue26789
        # also see: https://pythondev.readthedocs.io/finalization.html
        if sys.is_finalizing():
            raise RuntimeError(SYSTEM_FINALIZING_NO_GODOT_RERAISE)

    is_disallowed_error_logged = False
    for l in log_lines:
        if "ERROR" in l and not any(x in l for x in ERROR_ALLOWLIST):
            is_disallowed_error_logged = True
    return log_lines, is_disallowed_error_logged


def _raise_godot_error(
    log_path: str,
    artifact_path: Path,
    log_content: Optional[str] = None,
    details: str = "",
):
    if log_content is None:
        try:
            with open(log_path, "r") as infile:
                log_content = infile.read()
        except FileNotFoundError as e:
            log_content = str(e)

    if details != "":
        details = details + " "
    else:
        match = re.search(r"^ERROR: (.*)$", log_content, re.MULTILINE)
        if match is not None:
            details = match.group(1) + " "

    error_message = f"{details}\nLog: {log_path}\nArtifacts: {artifact_path}"

    logger.error(f"Godot error: {error_message}")
    logger.error("Attempting to dump log file here...")
    logger.error(log_content)

    logger.error("Attempting to log error to sentry...")
    log_to_sentry(GodotError(error_message))

    raise GodotError(error_message)


def _popen_process_group(bash_cmd: Iterable[str], log_path: str, env: Mapping[str, str] = {}):
    with open(log_path, "wb") as log:
        return subprocess.Popen(
            list(bash_cmd),
            start_new_session=True,
            stdout=log,
            stderr=log,
            env=env,
        )


def _kill_process_group(process: subprocess.Popen, is_lookup_error_ok: bool = True):
    try:
        pgid = os.getpgid(process.pid)
    except ProcessLookupError:
        if not is_lookup_error_ok:
            raise
        return
    assert pgid == process.pid, f"Cannot _kill_process_group of a child process (pgid={pgid}, pid={process.pid})"
    return os.killpg(pgid, SIGKILL)


def get_first_run_action_record_path(config_path: str) -> Path:
    return Path(config_path).parent.joinpath(f"actions.out")


def _derive_pipe_paths(config_path: str, log_path: str):
    action_record_path = get_first_run_action_record_path(config_path)
    if action_record_path.is_file():
        action_record_path = Path(log_path).parent.joinpath(_ACTION_REPLAY_FILENAME)
    action_pipe_path = log_path + ".actions"
    observation_pipe_path = log_path + ".observations"
    return str(action_record_path), action_pipe_path, observation_pipe_path


def create_error_artifact_path() -> Path:
    root_path = Path(GODOT_ERROR_LOG_PATH)
    root_path.mkdir(parents=True, exist_ok=True)
    return root_path / f"godot_env_artifacts__{str(uuid.uuid4())}.tar.gz"


class InteractiveGodotProcess:
    def __init__(
        self,
        config: SimSpec,
        gpu_id: int,
        keep_log: bool = True,
        is_dev_flag_added: bool = False,
        run_uuid: Optional[str] = None,
    ) -> None:
        self.gpu_id = gpu_id
        self.config = config
        self.keep_log = keep_log
        self.is_dev_flag_added = is_dev_flag_added
        self.process: Optional[subprocess.Popen[bytes]] = None
        self.log_uuid = str(uuid.uuid4())
        log_root = os.path.join(TEMP_DIR, self.log_uuid)
        os.makedirs(log_root, exist_ok=True)
        self.log_path = os.path.join(log_root, f"godot.log")
        self.prefix = f"Worker (PID={os.getpid()}): "

        if run_uuid is None:
            self.run_uuid = str(uuid.uuid4())
        else:
            self.run_uuid = run_uuid

        self.action_record_path, self.action_pipe_path, self.observation_pipe_path = _derive_pipe_paths(
            self.config_path, self.log_path
        )

        self.output_file_read_loc = 0
        self.error_marker = "ERROR"
        self.output_file_buffer = (" " * (len(self.error_marker) + 1)).encode("UTF-8")
        self.artifact_path = create_error_artifact_path()

    @property
    def config_path(self) -> str:
        return os.path.join(self.config.get_dir_root(), str(self.run_uuid), "config.json")

    @property
    def _config_done_path(self) -> str:
        return self.config_path.replace(".json", ".done")

    @property
    def is_replay(self):
        return basename(self.action_record_path) == _ACTION_REPLAY_FILENAME

    @property
    def is_running(self) -> bool:
        return self.process is not None and self.process.returncode is None

    @property
    def is_finished(self) -> bool:
        return self.process is not None and self.process.returncode is not None

    @property
    def is_closed(self) -> bool:
        if not self.is_finished:
            return False
        is_config_still_present = Path(self.config_path).is_file()
        does_config_done_exist = Path(self._config_done_path).is_file()

        config_has_been_renamed = not is_config_still_present and does_config_done_exist
        if config_has_been_renamed:
            return config_has_been_renamed

        has_been_cleaned_up = not (is_config_still_present or does_config_done_exist)
        if has_been_cleaned_up:
            return True
        return False

    def _read_log(self):
        return _read_log(self.log_path)

    def _get_godot_command(
        self,
        input_pipe_path: str,
        output_pipe_path: str,
        extra_flags: Tuple[str, ...] = tuple(),
    ) -> Tuple[List[str], Mapping[str, str]]:
        assert self.gpu_id is not None, "Refusing to run godot process without a gpu_id"
        extra_flags = extra_flags + tuple([f"--cuda-gpu-id={self.gpu_id}"])

        resolution = (self.config.recording_options.resolution_x, self.config.recording_options.resolution_y)
        if self.config.recording_options.is_adding_debugging_views:
            resolution = (resolution[0] * 2, resolution[1] * 2)

        bash_args = [
            DATAGEN_SCRIPT_PATH,
            f"--thread_count=4",
            f"-U",
            f"--input_pipe_path={input_pipe_path}",
            f"--output_pipe_path={output_pipe_path}",
            f"--resolution={resolution[0]}x{resolution[1]}",
            *extra_flags,
            self.config_path,
        ]
        bash_env = {
            GODOT_BINARY_PATH_ENV_FLAG: GODOT_BINARY_PATH,
        }
        return bash_args, bash_env

    def start(self) -> None:
        assert os.path.exists(GODOT_BINARY_PATH), (
            f"Cannot run avalon: Godot binary has not been installed to to {GODOT_BINARY_PATH_ENV_FLAG}={GODOT_BINARY_PATH}. "
            f"Please run `python -m avalon.install_godot_binary` and try again."
        )
        # we open a file here so that we can watch
        # create the file because we want to open it now
        Path(self.log_path).touch()
        self.output_file = open(self.log_path, "rb")

        Path(self.config_path).parent.mkdir(parents=True, exist_ok=True)
        with open(self.config_path, "w") as config_file:
            config_file.write(_create_json_from_config(self.config))

        extra_flags = ("--dev",) if self.is_dev_flag_added else tuple()
        godot_bash_args, godot_bash_env = self._get_godot_command(
            self.action_pipe_path, self.observation_pipe_path, extra_flags
        )
        logger.debug(f"{self.prefix} process group: {' '.join(godot_bash_args)}' &>> {self.log_path}")

        debug_bash_cmd = " ".join(self._get_godot_command(self.action_record_path, "/tmp/debug_output")[0])
        logger.debug(f"{self.prefix} TO DEBUG RUN: {debug_bash_cmd}")

        self.process = _popen_process_group(godot_bash_args, self.log_path, godot_bash_env)

    def save_artifacts(self, recent_worlds: Deque[GenerateAvalonWorldParams]) -> Path:
        tar_path = Path(self.artifact_path)
        with tarfile.open(tar_path, "x:gz") as tar:
            tar.add(self.action_record_path, arcname=Path(self.action_record_path).name)
            tar.add(self.log_path, arcname=Path(self.log_path).name)
            tar.add(self.config_path, arcname=Path(self.config_path).name)

            generated_root_path_by_id = {}
            for i, world in enumerate(recent_worlds):
                generation_id = uuid.uuid4()
                for world_path in world.output_path.rglob("*.tscn"):
                    generated_root_path_by_id[str(generation_id)] = str(world.output_path)
                    tar.add(world_path, arcname=f"worlds/{generation_id}/{world_path.name}")

            with tempfile.NamedTemporaryFile(suffix=".json", mode="w+") as tmp_file:
                json.dump(
                    {"run_uuid": str(self.run_uuid), "generated_root_path_by_id": generated_root_path_by_id},
                    tmp_file,
                )
                tmp_file.flush()
                tar.add(tmp_file.name, arcname=f"meta.json")
        return tar_path

    def check_for_errors(self) -> None:
        """Raises a GodotError if an error was encountered"""
        file_size = os.path.getsize(self.log_path)
        unread_byte_count = file_size - self.output_file_read_loc
        if unread_byte_count > 0:
            new_bytes = self.output_file.read(unread_byte_count)
            self.output_file_buffer = self.output_file_buffer + new_bytes
            line_buffer = self.output_file_buffer.decode("UTF-8", errors="ignore")
            lines = line_buffer.split("\n")
            # NOTE: we may have dropped something due to the "ignore" above but that should be fine
            self.output_file_buffer = lines.pop().encode("UTF-8")
            for line in lines:
                if self.error_marker in line and not any(x in line for x in ERROR_ALLOWLIST):
                    self._raise_error()

    def close(self, kill: bool = True, raise_logged_errors: bool = True):
        assert not self.is_closed, f"{self} already closed, cannot close again"

        self.output_file.close()

        if self.process is None:
            return

        if kill:
            _kill_process_group(self.process)

        if raise_logged_errors:
            self.raise_any_logged_godot_errors()

        try:
            wait_until_true(self._poll_for_exit)
        except Exception as e:
            _kill_process_group(self.process)
            if raise_logged_errors:
                raise
            logger.error(f"Got error {e} but continuing to exit")

    def _error_code_repr(self) -> str:
        if self.process is None or not self.process.returncode:
            return ""
        signal = abs(self.process.returncode)
        matching_known_signal = (s.name for s in valid_signals() if isinstance(s, Signals) and s.value == signal)
        name = next(matching_known_signal, "UNKNOWN")
        return f"returncode={name}({self.process.returncode})"

    def _raise_error(self, log_content: Optional[str] = None) -> None:
        _raise_godot_error(
            log_path=self.log_path,
            artifact_path=self.artifact_path,
            log_content=log_content,
            details=self._error_code_repr(),
        )

    def raise_any_logged_godot_errors(self) -> None:
        lines, is_error_logged = _read_log(self.log_path)
        if is_error_logged:
            self._raise_error("\n".join(lines))

    def wait_for_log_signal(self, log_signal: str) -> None:
        process = self.process
        assert process is not None, "Cannot wait_for_ready_signal() before start()"

        def wait_for_ready() -> bool:
            assert process is not None
            process.poll()
            if process.returncode:
                self._raise_error()
            log_lines, is_error_logged = self._read_log()
            if is_error_logged:
                _raise_godot_error(
                    log_path=self.log_path,
                    artifact_path=self.artifact_path,
                    log_content="\n".join(log_lines),
                )
            return any(log_signal in l for l in log_lines)

        try:
            wait_until_true(wait_for_ready, sleep_inc=0.0001, max_wait_sec=20)
        except TimeoutError as e:
            err = f"did not observe ready message {log_signal}: {str(e)}"
            _raise_godot_error(log_path=self.log_path, artifact_path=self.artifact_path, details=err)

    def wait_for_ready_signal(self) -> None:
        self.wait_for_log_signal(READY_LOG_SIGNAL)

    def _poll_for_exit(self) -> bool:
        assert self.process is not None, "Cannot call _poll_for_exit() before start()"
        self.raise_any_logged_godot_errors()
        self.process.poll()
        if self.process.returncode is None:
            return False
        self.process.wait()
        if self.process.returncode:
            self._raise_error()

        # TODO: this whole mess is not multiprocess-safe
        os.rename(self.config_path, self._config_done_path)
        if not self.keep_log:
            if os.path.exists(self.log_path):
                os.remove(self.log_path)

        return True



from typing import Callable
from typing import ClassVar
from typing import Dict
from typing import Optional
from typing import Tuple
from typing import Type
from typing import TypeVar
from typing import cast

import attr
import numpy as np
from godot_parser import GDObject
from godot_parser import Node as GDNode
from scipy.spatial.transform import Rotation

from avalon.common.utils import to_immutable_array
from avalon.datagen.world_creation.constants import IDENTITY_BASIS
from avalon.datagen.world_creation.entities.constants import CANONICAL_FOOD_HEIGHT_ON_TREE
from avalon.datagen.world_creation.entities.constants import FOOD_TREE_BASE_HEIGHT
from avalon.datagen.world_creation.entities.item import InstancedDynamicItem
from avalon.datagen.world_creation.entities.tools.tool import Tool
from avalon.datagen.world_creation.entities.tools.weapons import LargeRock
from avalon.datagen.world_creation.entities.tools.weapons import Rock
from avalon.datagen.world_creation.entities.tools.weapons import Stick
from avalon.datagen.world_creation.entities.utils import facing_2d
from avalon.datagen.world_creation.entities.utils import get_random_ground_points
from avalon.datagen.world_creation.entities.utils import normalized
from avalon.datagen.world_creation.types import GodotScene
from avalon.datagen.world_creation.types import MapBoolNP
from avalon.datagen.world_creation.types import Point3DListNP
from avalon.datagen.world_creation.types import Point3DNP
from avalon.datagen.world_creation.utils import scale_basis
from avalon.datagen.world_creation.worlds.height_map import HeightMap

_FoodT = TypeVar("_FoodT", bound="Food")


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Food(InstancedDynamicItem):
    mass: float = 1.0
    resource_file: str = "res://items/food.tscn"
    energy: float = 1.0
    plant_path: Optional[str] = None

    position: np.ndarray = attr.ib(
        default=np.array([0.0, 0.0, 0.0]),
        converter=to_immutable_array,
        eq=attr.cmp_using(eq=np.array_equal),  # type: ignore[attr-defined]
    )
    scale: np.ndarray = np.array([1.0, 1.0, 1.0])
    is_grown_on_trees: bool = True
    is_found_on_ground: bool = True
    is_openable: bool = False
    count_distribution: Tuple[Tuple[float, ...], Tuple[float, ...]] = ((0.0, 1.0), (1.0, 1.7))

    @property
    def additional_offset(self) -> np.ndarray:
        return np.array([0.0, 0.0, 0.0])

    def get_stem_joint(self, food_name: str) -> GDNode:
        basis = scale_basis(IDENTITY_BASIS, self.scale)
        point_five_over_parent = GDObject("Transform", *basis, 0, 0.5 * self.scale[1], 0)
        return GDNode(
            f"{food_name}_stem_joint",
            type="Generic6DOFJoint",
            properties={
                "transform": point_five_over_parent,
                "nodes/node_a": f"../../{food_name}",
                "nodes/node_b": f"../../{self.plant_path}",
                "linear_limit_x/upper_distance": 0.5,
                "linear_limit_y/upper_distance": 0.0,
                "linear_limit_z/upper_distance": 0.5,
            },
        )

    def attached_to(self: _FoodT, tree: "FoodTree") -> _FoodT:
        assert tree.entity_id > -1, f"Food must be attached_to the result of world.add_item({tree})"
        assert tree.is_food_on_tree, f"Food shouldn't be attached_too tree {tree} without is_food_on_tree=True"
        return attr.evolve(self, plant_path=tree.node_name)

    def get_node(self, scene: GodotScene) -> GDNode:
        food_node = super().get_node(scene)
        food_node.properties["energy"] = self.energy

        if self.plant_path is None:
            return food_node

        food_node.add_child(self.get_stem_joint(food_node.name))
        return food_node

    def get_count(self, rand: np.random.Generator) -> int:
        return round(float(np.interp(rand.uniform(), self.count_distribution[0], self.count_distribution[1])))

    def get_opened_version(self) -> "Food":
        if not self.is_openable:
            return self
        raise NotImplementedError("Openable food must implement get_opened_version")

    def get_tool_options(self) -> Tuple[float, Tuple[Type[Tool], ...]]:
        return 0, tuple()

    @property
    def is_tool_required(self) -> bool:
        return self.get_tool_options()[0] == 1

    @property
    def is_tool_useful(self) -> bool:
        return self.get_tool_options()[0] > 0

    def is_always_multiple(self) -> bool:
        return False


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Banana(Food):
    resource_file: str = "res://items/food/banana.tscn"
    energy: float = 0.5

    @property
    def additional_offset(self) -> np.ndarray:
        if self.plant_path is not None:
            return np.array([0.0, 0.3, 0.0])
        return super().additional_offset


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Apple(Food):
    resource_file: str = "res://items/food/apple.tscn"


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Fig(Food):
    resource_file: str = "res://items/food/fig.tscn"


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class OpenableFood(Food):
    is_openable: bool = True

    @property
    def is_open(self) -> bool:
        return not self.is_openable

    def get_opened_version(self: "OpenableFood") -> "OpenableFood":
        if self.is_open:
            return self
        opened: "OpenableFood"
        with self.mutable_clone() as opened:
            opened.resource_file = opened.resource_file.replace(".tscn", "_open.tscn")
            opened.is_openable = False
            return opened


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Orange(OpenableFood):
    resource_file: str = "res://items/food/orange.tscn"

    def get_tool_options(self):
        return 0.1, (Rock, Stick, LargeRock)


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Avocado(OpenableFood):
    resource_file: str = "res://items/food/avocado.tscn"

    def get_tool_options(self):
        return 1, (Rock, LargeRock)


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Coconut(OpenableFood):
    resource_file: str = "res://items/food/coconut.tscn"

    def get_tool_options(self):
        return 0.25, (Rock, Stick, LargeRock)


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Honeycomb(Food):
    resource_file: str = "res://items/food/honeycomb.tscn"
    is_found_on_ground: bool = False


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Cherry(Food):
    resource_file: str = "res://items/food/cherry.tscn"

    @property
    def additional_offset(self) -> np.ndarray:
        if self.plant_path is not None:
            return np.array([0.0, -0.5, 0.0])
        return super().additional_offset

    def is_always_multiple(self) -> bool:
        return True


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Mulberry(Food):
    resource_file: str = "res://items/food/mulberry.tscn"

    @property
    def additional_offset(self) -> np.ndarray:
        if self.plant_path is not None:
            return np.array([0.0, 0.3, 0.0])
        return super().additional_offset


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Carrot(Food):
    is_grown_on_trees: bool = False
    resource_file: str = "res://items/food/carrot.tscn"

    def get_offset(self) -> float:
        return 0.0


FOODS: Tuple[Food, ...] = (
    Apple(),
    Banana(),
    Cherry(),
    Honeycomb(),
    Mulberry(),
    Fig(),
    Orange(),
    Avocado(),
    Coconut(),
    Carrot(),
)
NON_TREE_FOODS = [x for x in FOODS if not x.is_grown_on_trees]
CANONICAL_FOOD = FOODS[0]
CANONICAL_FOOD_CLASS = FOODS[0].__class__
CANONICAL_FOOD_HEIGHT = CANONICAL_FOOD.get_offset()
GetHeightAt = Callable[[Tuple[float, float]], float]


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class FoodTreeBase(InstancedDynamicItem):
    resource_file: str = "res://scenery/tree_base.tscn"
    rotation: Rotation = Rotation.identity()

    def get_offset(self) -> float:
        return -0.25


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class FoodTree(InstancedDynamicItem):
    ROTATION_TO_SUITABLE_OFFSET: ClassVar[Dict[Tuple[float], Tuple[float, float]]] = {
        (0,): (1.25, CANONICAL_FOOD_HEIGHT_ON_TREE),
        # TODO add mechanism for multiple rotations
        # (45,): (0.875, 3.25),
    }

    @property
    def food_offset(self) -> Tuple[float, float]:
        return self.ROTATION_TO_SUITABLE_OFFSET[(0,)]

    def _relative_position(self, offset: Point3DNP) -> Point3DNP:
        return cast(Point3DNP, self.rotation.apply(offset) * self.scale)

    def get_food_height(self, food: _FoodT) -> float:
        xz, y = self.food_offset
        return cast(float, y * self.scale[1])

    resource_file: str = "res://scenery/trees/fruit_tree_normal.tscn"
    is_tool_helpful: bool = True
    scale: np.ndarray = np.array([1.0, 1.0, 1.0])
    rotation: Rotation = Rotation.identity()

    is_food_on_tree: bool = False

    @property
    def height(self):
        scale_y = self.scale[1]
        return FOOD_TREE_BASE_HEIGHT * scale_y

    @staticmethod
    def build(tree_height: float, is_food_on_tree: bool = False) -> "FoodTree":
        scale_factor = tree_height / FOOD_TREE_BASE_HEIGHT
        return FoodTree(
            scale=scale_factor * np.array([1.0, 1.0, 1.0]),
            position=np.array([0.0, 0.0, 0.0]),
            is_food_on_tree=is_food_on_tree,
        )

    def _resolve_trunk_placement(
        self, spawn: Point3DNP, primary_food: Point3DNP, get_height_at: GetHeightAt
    ) -> Tuple[Point3DNP, float]:
        goal_vector: Point3DNP = primary_food - spawn
        goal_vector[1] = 0
        goal_vector = normalized(goal_vector)
        if self.is_food_on_tree:
            xz_offset, y_offset = self.food_offset
            away_from_spawn_and_down_from_food = goal_vector * xz_offset
            away_from_spawn_and_down_from_food[1] = -y_offset
            away_from_spawn_and_down_from_food *= self.scale
            position = primary_food + away_from_spawn_and_down_from_food
            return position, get_height_at((position[0], position[2]))

        # TODO: could use a bit more variety in how the tree is place relative to the food
        x, _, z = primary_food + 2.0 * goal_vector
        ground_level = get_height_at((x, z))
        ground_level_behind_food = np.array([x, ground_level, z])
        return ground_level_behind_food, ground_level

    def place(
        self, spawn: Point3DNP, primary_food: Point3DNP, get_height_at: GetHeightAt
    ) -> Tuple["FoodTree", Optional[FoodTreeBase]]:
        trunk_placement, ground_height = self._resolve_trunk_placement(spawn, primary_food, get_height_at)
        rotation = facing_2d(trunk_placement, spawn)
        root_affordance = 0.25 * self.scale[1]
        is_boulder_needed = self.is_food_on_tree and ground_height < (trunk_placement[1] - root_affordance)
        tree = attr.evolve(self, position=trunk_placement, rotation=rotation)
        if not is_boulder_needed:
            return (tree, None)

        return tree, FoodTreeBase(position=trunk_placement, rotation=rotation)

    def get_food_locations(
        self,
        rand: np.random.Generator,
        center: Point3DNP,
        count: int,
        map: HeightMap,
        min_radius: float,
        max_radius: float,
        offset: float,
        island_mask: MapBoolNP,
    ) -> Point3DListNP:
        # we originally intended that some food could sometimes be at different places on the tree and some on the
        # ground, but didn't get to finish this feature before the deadline
        ground_count = count
        ground_points = get_random_ground_points(
            rand, center, ground_count, map, min_radius, max_radius, offset, island_mask
        )
        tree_point_count = count - ground_count
        if tree_point_count == 0:
            return ground_points
        random_heights = rand.uniform(0.5, 0.9, size=tree_point_count) * self.height
        zeros = np.zeros_like(random_heights)
        height_vectors = np.stack([zeros, random_heights, zeros])
        tree_points = self.position + height_vectors
        return cast(Point3DListNP, tree_points)

    def get_offset(self) -> float:
        return 0.0

    def get_node(self, scene: GodotScene) -> GDNode:
        node = super().get_node(scene)
        node.properties["scale"] = GDObject("Vector3", *self.scale)
        return node



import subprocess
from typing import Optional
from typing import Tuple


def load_aws_keys() -> Tuple[Optional[str], Optional[str]]:
    result = subprocess.run("cat ~/.aws/credentials", shell=True, capture_output=True)
    is_relevant_section = False
    access_key = None
    secret_key = None
    for line in result.stdout.decode("UTF-8").splitlines():
        if is_relevant_section:
            if access_key is not None and secret_key is not None:
                break
            elif line.startswith("aws_access_key_id = "):
                access_key = line.split(" ")[-1].strip()
            elif line.startswith("aws_secret_access_key = "):
                secret_key = line.split(" ")[-1].strip()
            elif line.strip() == "":
                pass
            else:
                raise Exception(f"Bad line in credentials file: {line}")
        if "[computronium]" in line:
            is_relevant_section = True
    return access_key, secret_key



from typing import List
from typing import Optional

import numpy as np
from scipy.spatial.transform import Rotation
from trimesh import Trimesh
from trimesh.visual import ColorVisuals

from avalon.datagen.world_creation.types import Point3DNP
from avalon.datagen.world_creation.types import RGBATuple


def homogeneous_transform_matrix(
    position: Point3DNP = np.array([0, 0, 0]), rotation: Optional[Rotation] = None
) -> np.ndarray:
    if rotation is None:
        rotation = np.eye(3)
    else:
        rotation = rotation.as_matrix()
    transform = np.eye(4)
    transform[:3, :3] = rotation
    transform[:3, 3] = position
    return transform


def make_color_visuals(mesh: Trimesh, rgba: RGBATuple) -> ColorVisuals:
    return ColorVisuals(mesh, face_colors=np.repeat([rgba], len(mesh.faces), axis=0))


class MeshData:
    def __init__(self) -> None:
        self.vertices: List[List[float]] = []
        self.faces: List[List[float]] = []
        self.face_normals: List[List[float]] = []
        self.face_colors: List[float] = []
        self.index_offset = 0


def unnormalize(mesh: Trimesh) -> Trimesh:
    """Re-create mesh such that none of its faces share vertices"""
    points_per_face = 3
    coords_per_point = 3
    face_count = len(mesh.faces)
    new_vertices = np.empty((face_count * points_per_face, coords_per_point), dtype=np.float32)
    for i, face in enumerate(mesh.faces):
        offset = i * points_per_face
        new_vertices[offset : offset + points_per_face, :] = mesh.vertices[face]
    new_faces = np.array(range(len(new_vertices))).reshape(-1, coords_per_point)
    return Trimesh(vertices=new_vertices, faces=new_faces, process=False)



import attr
import godot_parser
import numpy as np
from godot_parser import GDObject
from godot_parser import Node as GDNode

from avalon.datagen.world_creation.entities.tools.tool import Tool
from avalon.datagen.world_creation.types import GodotScene
from avalon.datagen.world_creation.utils import make_transform
from avalon.datagen.world_creation.utils import scale_basis


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Stone(Tool):
    size: float = 1.0
    mass: float = 10.0
    resource_file: str = "res://items/stone.tscn"
    script_file: str = "res://items/stone.gd"
    default_mesh_extents: float = 1.0

    def get_node(self, scene: GodotScene) -> GDNode:
        mesh_node = super().get_node(scene)
        item_name = mesh_node.name
        mesh_node.name = "mesh"

        scale_factor = self.size / self.default_mesh_extents
        mesh_node.properties["transform"] = make_transform(
            scale_basis(np.array([1, 0, 0, 0, 1, 0, 0, 0, 1], dtype=np.float64), np.array([scale_factor] * 3))
        )

        half_size = self.size / 2
        collision_shape_node = GDNode(
            "collision_shape",
            type="CollisionShape",
            properties={
                "shape": scene.add_sub_resource(
                    "BoxShape",
                    extents=godot_parser.Vector3(half_size, half_size, half_size),
                ).reference,
            },
        )
        item_node = GDNode(
            item_name,
            type="RigidBody",
            properties={
                "transform": GDObject("Transform", *self.basis, *self.position),
                "script": scene.add_ext_resource(self.script_file, "Script").reference,
                "entity_id": self.entity_id,
                "mass": self.mass,
            },
        )
        item_node.add_child(collision_shape_node)
        item_node.add_child(mesh_node)
        return item_node

    def get_offset(self) -> float:
        return (self.size * self.default_mesh_extents) / 2.0



import attr

from avalon.contrib.serialization import Serializable


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class WorldConfig(Serializable):
    # random number. Used by some other functions to create a deterministic source of randomness so that we can
    # guarantee that the same worlds are produced every time the function is called with the same seed.
    seed: int
    # how large the world is, in meters, along one edge
    size_in_meters: float
    # how many points will be created for triangulation of the heightmap.
    # controls how detailed the terrain is
    # You probably should not change this unless you plan on re-tuning and changing quite a few other constants, as
    # it affects how reliably distances can be specified in tasks (ex: if there are very few points per meter, it is
    # difficult to control, for example, the exact size of a gap over which we want the agent to jump).
    # This parameter is a trade-off between how fine our control is over the geometry of the world and performance.
    # (at any given world size)
    point_density_in_points_per_square_meter: float
    # terrain construction works by iteratively subdividing a grid.
    # this controls the initial size of the grid.
    # lower numbers are much higher diversity terrain because the iteration works by adding noise, then subdividing.
    # Min is 3 because we need to be able to interpolate
    initial_point_count: int
    # TODO: DEPRECATED:
    # the scale of the noise actually doesn't matter because we reset the height afterwards
    # only really matters relative to mountain_noise_scale
    initial_noise_scale: float
    # how many times to do the add noise + subdivide loop
    fractal_iteration_count: int
    # how quickly the terrain noise decays with each iteration.
    # should be set to < 1.0 to make the noise get smaller each iteration (as it is multiplied by the noise)
    # setting it > 1.0 is allowed, though it will result if crazy jagged worlds that are not particularly interesting.
    noise_scale_decay: float
    # how many "mountains" to create. Mountains, in our terrain generation code, are simply mounds of height that are
    # added to the terrain. This factor controls how many of them to add at each iteration.
    # setting to zero will turn off mountains completely, making the rest of the mountain parameters pointless
    mountain_noise_count: int
    # how large the mound is that is created
    mountain_radius: float
    # how far away from the edge of the map to stay. Helpful for keeping the mountains towards the center to ensure
    # that there will be some land
    mountain_offset: float
    # how the mountain radius is updated with each fractal iteration. Same decay logic as noise_scale_decay
    mountain_radius_decay: float
    # how large the mountain effect is. Note that this only matters relative to the initial_noise_scale, which we
    # canonically set to 1.0.  Setting mountain_noise_scale to larger values means that more of the noise will come
    # from the mountain effect vs the random noise effect.
    mountain_noise_scale: float
    # whether mountain placement is normally or uniformly distributed
    is_mountain_placement_normal_distribution: bool
    # how high to make the highest point of the map (in meters above sea level).
    final_max_altitude_meters: float
    # this is the fraction of the map on each edge that will be linearly interpolated towards zero to ensure that
    # the world is a well-formed island (ie, its edge intersects the ocean)
    final_world_edge_fade_fraction: float = 0.05
    # unlike the above parameter, this controls a sort of fading/lowering of height that does not guarantee that the
    # terrain will reach the ocean. However, it certainly helps in most cases, and it looks much more natural
    # this parameter controls the fraction of the world, on each edge, over which this effect is applied
    fade_fraction: float = 0.2
    # the spatial scale of the noise that is applied to this fading effect. Smaller values make bigger patterns
    fade_noise_scale: float = 0.01
    # how harsh to make the fading effect. Smaller values will make it have a larger effect
    fade_noise_min: float = 0.5
    # whether to apply this fading effect in a circular or square way. The circlse look nicer
    is_fade_circular: bool = True
    # if greater than zero, controls the standard deviation of the blur kernel that is applied
    blur_meters_std_dev: float = 0
    # is set to true on maps where we create a world, but it's pointless, because the only interesting feature is a
    # building, which you spawn inside
    is_indoor_only: bool = False



from __future__ import annotations

from collections import defaultdict
from typing import Tuple
from typing import Type

import attr
import gym
import numpy as np
import torch
import torch.nn.functional as F
from einops import rearrange
from torch import Tensor
from tree import map_structure

from avalon.agent.common import wandb_lib
from avalon.agent.common.action_model import DictActionDist
from avalon.agent.common.action_model import visualize_action_dists
from avalon.agent.common.types import ActionBatch
from avalon.agent.common.types import Algorithm
from avalon.agent.common.types import ObservationBatch
from avalon.agent.common.util import explained_variance
from avalon.agent.ppo.gae import gae
from avalon.agent.ppo.model import CNNBase
from avalon.agent.ppo.model import MLPBase
from avalon.agent.ppo.observation_model import ObservationModel
from avalon.agent.ppo.observation_model import PPOModel
from avalon.agent.ppo.params import PPOParams
from avalon.agent.ppo.ppo_types import PPOBatchDataWithGAE
from avalon.agent.ppo.ppo_types import PPOBatchSequenceDataWithGAE
from avalon.agent.ppo.ppo_types import PPOStepData
from avalon.contrib.utils import set_all_seeds


class PPO(Algorithm["PPOParams"]):
    step_data_type: Type = PPOStepData

    def __init__(self, params: PPOParams, observation_space: gym.spaces.Dict, action_space: gym.spaces.Dict) -> None:
        self.params = params
        super().__init__(params, observation_space, action_space)

        self.model: PPOModel
        if params.model == "cnn":
            self.model = CNNBase(params, observation_space, action_space, hidden_size=64)
        elif params.model == "mlp":
            self.model = MLPBase(params, observation_space, action_space)
        elif params.model == "godot":
            self.model = ObservationModel(params, observation_space, action_space)
        else:
            assert False

        self.optim = torch.optim.Adam(self.model.parameters(), lr=params.lr, eps=1e-5)
        self.seed_state = defaultdict(int)

    def forward(self, obs: ObservationBatch):
        return self.model(obs)

    def rollout_step(
        self,
        next_obs: ObservationBatch,
        dones: Tensor,  # shape (batch_size, )
        indices_to_run: list[bool],  # shape (batch_size, )
        exploration_mode: str,
    ) -> Tuple[ActionBatch, dict]:
        step_values: Tensor
        dist: DictActionDist
        step_values, dist = self.model(next_obs)

        # Sample actions from the policy distribution
        # This should be a list (of len num_workers) of action dicts
        step_actions = dist.sample()

        if self.params.deterministic:
            # Determinism here is tricky, since we allow partial batches.
            # Need to do a separate sample for each worker, setting the seed before each,
            # otherwise the random "seed" for each sample will differ depending on ordering.
            # This will overwrite the previous action sample, to avoid having to construct the arrays.
            for i, should_run in enumerate(indices_to_run):
                # We do need a new seed for each sample,
                # otherwise we always sample the same action as long as the probs are roughly uniform
                # (it's not a chaotic sampling algorithm).
                # So we construct one based on worker_id that gets incremented in a determinstic way
                # irrelevant of the ordering of partial batches.
                if should_run:
                    if self.seed_state[i] == 0:
                        self.seed_state[i] = i * 100000
                    set_all_seeds(self.seed_state[i])

                    def assign(target, source) -> None:
                        target[i] = source

                    map_structure(assign, step_actions, dist[i].sample())
                    self.seed_state[i] += 1

        step_policy_probs = dist.log_prob(step_actions)
        policy_entropy = dist.entropy().cpu()

        # Store data for use in training
        step_actions = map_structure(lambda x: x.detach().cpu(), step_actions)

        step_values = step_values.detach().cpu()
        step_policy_probs = step_policy_probs.detach().cpu()
        to_store = {"value": step_values, "policy_prob": step_policy_probs, "policy_entropy": policy_entropy}
        return step_actions, to_store

    def train_step(self, rollouts: PPOBatchSequenceData, i: int) -> int:  # type: ignore
        # the observations have an extra frame at the end to use for value backup.
        final_observations: ObservationBatch = {k: v[:, -1] for k, v in rollouts.observation.items()}
        rollouts = attr.evolve(rollouts, observation=map_structure(lambda x: x[:, :-1], rollouts.observation))

        if self.params.entropy_mode == "max" and self.params.entropy_coef != 0:
            entropy_reward: Tensor = rollouts.policy_entropy * self.params.entropy_coef
            wandb_lib.log_histogram("training/entropy_reward", entropy_reward, i)
            wandb_lib.log_histogram("training/env_reward", rollouts.reward, i)
            rollouts = attr.evolve(rollouts, reward=rollouts.reward + entropy_reward)

        with torch.no_grad():
            final_values = self.model(final_observations)[0].detach()

            # Compute advantages and future discounted rewards with GAE
            advantages = gae(
                rollouts.reward, rollouts.value, rollouts.done, self.params.discount, self.params.lam, final_values
            )
            rewards_to_go = advantages + rollouts.value
            rollouts = PPOBatchSequenceDataWithGAE(
                **attr.asdict(rollouts), advantage=advantages, reward_to_go=rewards_to_go
            )

            # flatten (num_workers,num_steps,...) into ((num_workers*num_steps,...)
            train_data: PPOBatchDataWithGAE = map_structure(lambda x: rearrange(x, "b t ... -> (b t) ..."), rollouts)

        for epoch in range(self.params.ppo_epochs):
            indices = np.random.permutation(self.params.num_steps * self.params.num_workers)

            for batch in range(self.params.num_batches):
                start = batch * self.params.batch_size
                end = start + self.params.batch_size

                batch_indices = indices[start:end]
                batch_data = map_structure(lambda x: x[batch_indices], train_data)
                self.train_substep(batch_data, i)

                i += 1
            assert end == self.params.num_steps * self.params.num_workers

        # Lots of logging!
        ev = explained_variance(rollouts.value.detach().flatten().cpu(), rewards_to_go.detach().flatten().cpu())
        wandb_lib.log_scalar("training/rollout_value_ev", ev, i)
        wandb_lib.log_histogram("training/raw_advantage", advantages, i)
        wandb_lib.log_histogram("training/rewards", rollouts.reward, i)

        return i

    def train_substep(self, batch: PPOBatchDataWithGAE, i: int) -> None:
        self.model.train()
        self.optim.zero_grad()

        # Normalize the advantages.
        # Note the advantages are normalized by batch, not the entire epoch altogether
        # might want to parameterize the constant here. I had it at 1e-5 originally
        advantages = (batch.advantage - batch.advantage.mean()) / (batch.advantage.std() + 1e-8)

        values_new, dist_new = self.model(batch.observation)
        selected_prob_new = dist_new.log_prob(batch.action)

        # Compute the PPO loss
        prob_ratio = torch.exp(selected_prob_new - batch.policy_prob)
        a = prob_ratio * advantages
        b = torch.clamp(prob_ratio, 1 - self.params.clip_range, 1 + self.params.clip_range) * advantages
        ppo_loss = -1 * torch.min(a, b)
        ppo_loss = ppo_loss.mean()

        # Compute the value function loss
        # Clipped loss - same idea as PPO loss, don't allow value to move too
        # far from where it was previously
        if self.params.clip_range_vf is None:
            value_pred_clipped = values_new
        else:
            value_pred_clipped = batch.value + (values_new - batch.value).clamp(
                -self.params.clip_range_vf, self.params.clip_range_vf
            )

        # NOTE: implementations differ in how they handle this.
        if self.params.baselines_style_vf_loss:
            # this is how it's done in baselines ppo2
            value_losses = (values_new - batch.reward_to_go) ** 2
            value_losses_clipped = (value_pred_clipped - batch.reward_to_go) ** 2
            value_loss = 0.5 * torch.max(value_losses, value_losses_clipped)
            value_loss = value_loss.mean()
        else:
            # but sb3 uses a simpler method:
            value_loss = F.mse_loss(batch.reward_to_go, value_pred_clipped)
        policy_entropy = torch.mean(dist_new.entropy())

        base_loss = ppo_loss + self.params.value_loss_coef * value_loss
        if self.params.entropy_mode == "regularized" and self.params.entropy_coef != 0:
            entropy_loss = -1 * self.params.entropy_coef * policy_entropy
            loss = base_loss + entropy_loss
            wandb_lib.log_scalar("loss/entropy_loss", entropy_loss, i)
        else:
            loss = base_loss

        loss.backward()
        torch.nn.utils.clip_grad_norm_(self.model.parameters(), self.params.clip_grad_norm)
        self.optim.step()

        # Lots of logging
        clip_fraction = torch.mean((torch.abs(prob_ratio - 1) > self.params.clip_range).float()).item()
        wandb_lib.log_scalar("loss/clip_fraction", clip_fraction, i)
        wandb_lib.log_histogram("loss/ppo_loss", -1 * torch.min(a, b), i)
        wandb_lib.log_scalar("loss/value_loss", value_loss, i)
        wandb_lib.log_scalar("loss/loss", loss, i)
        wandb_lib.log_histogram("training/value_pred", values_new, i)
        wandb_lib.log_histogram("training/advantage", advantages, i)
        wandb_lib.log_histogram("training/value_target", batch.reward_to_go, i)
        wandb_lib.log_histogram("training/prob_ratio", prob_ratio, i)
        wandb_lib.log_scalar("policy/policy entropy", policy_entropy, i)
        # Approximate KL divergence between the policy when it was rolled out and now.
        # Explained in https://iclr-blog-track.github.io/2022/03/25/ppo-implementation-details/ and
        # http://joschu.net/blog/kl-approx.html
        approxkl = 0.5 * ((batch.policy_prob - selected_prob_new) ** 2).mean()
        wandb_lib.log_scalar("policy/approx_kl", approxkl, i)

        # Log actions
        visualize_action_dists(self.action_space, dist_new, prefix="train_policy")



from pathlib import Path
from typing import Optional
from typing import Tuple

import attr
import numpy as np
from scipy import stats

from avalon.datagen.errors import WorldTooSmall
from avalon.datagen.world_creation.configs.export import ExportConfig
from avalon.datagen.world_creation.configs.task import TaskConfig
from avalon.datagen.world_creation.constants import CLIMBING_REQUIRED_HEIGHT
from avalon.datagen.world_creation.constants import TIGHT_DIST_STD_DEV
from avalon.datagen.world_creation.constants import AvalonTask
from avalon.datagen.world_creation.tasks.eat import add_food_tree_for_simple_task
from avalon.datagen.world_creation.worlds.creation import create_world_from_constraint
from avalon.datagen.world_creation.worlds.difficulty import difficulty_variation
from avalon.datagen.world_creation.worlds.difficulty import normal_distrib_range
from avalon.datagen.world_creation.worlds.difficulty import scale_with_difficulty
from avalon.datagen.world_creation.worlds.difficulty import select_boolean_difficulty
from avalon.datagen.world_creation.worlds.export import export_world
from avalon.datagen.world_creation.worlds.obstacles.configure import make_ring
from avalon.datagen.world_creation.worlds.obstacles.height_path import HeightPath
from avalon.datagen.world_creation.worlds.obstacles.height_solution import HeightSolution
from avalon.datagen.world_creation.worlds.types import CompositionalConstraint
from avalon.datagen.world_creation.worlds.world import World
from avalon.datagen.world_creation.worlds.world_locations import WorldLocations

_MIN_GAP_DISTANCE = 3.0


@attr.s(auto_attribs=True, collect_by_mro=True, hash=True)
class MoveTaskConfig(TaskConfig):
    # move has two different modes: direct and indirect.
    # in the direct mode, it's just a normal world and you walk somewhere, very straightforward
    # in the indirect mode, there is a very particular path that must be followed in order to get to the destination
    # these values control how far away the goal (food) is from the starting point in each of the two modes
    direct_goal_dist_easy: float = 0.5
    direct_goal_dist_hard: float = 40.0
    indirect_goal_dist_easy: float = 6.0
    indirect_goal_dist_hard: float = 40.0
    # these parameters control how many turns there are in the path (for the indirect mode)
    # due to implementation constraints, it cannot be greater than 2, but we dont want to be making weird mazes anyway
    indirect_path_point_count_easy: int = 0
    indirect_path_point_count_hard: int = 2
    indirect_path_point_count_std_dev: float = 0.25
    # controls how wide the path is. As this becomes more narrow, the task gets closer to walking a tightrope
    indirect_path_width_easy: float = 10.0
    indirect_path_width_hard: float = 1.75
    indirect_path_width_std_dev: float = 0.1
    # how deep to make the chasm that must be crossed in indirect mode. Should be deep enough that you can't just jump
    # out of it, otherwise won't learn to actually walk along the path
    indirect_chasm_depth_easy: float = CLIMBING_REQUIRED_HEIGHT
    indirect_chasm_depth_hard: float = CLIMBING_REQUIRED_HEIGHT * 3.0


def generate_move_task(
    rand: np.random.Generator,
    difficulty: float,
    output_path: Path,
    export_config: ExportConfig,
    task_config: MoveTaskConfig = MoveTaskConfig(),
) -> None:
    world, locations, difficulty = create_move_obstacle(rand, difficulty, export_config, task_config=task_config)
    world, locations = world.end_height_obstacles(locations, is_accessible_from_water=False)
    world = add_food_tree_for_simple_task(world, locations)
    world = world.add_spawn(rand, difficulty, locations.spawn, locations.goal)
    export_world(output_path, rand, world)


def create_move_obstacle(
    rand: np.random.Generator,
    difficulty: float,
    export_config: ExportConfig,
    constraint: Optional[CompositionalConstraint] = None,
    task_config: MoveTaskConfig = MoveTaskConfig(),
) -> Tuple[World, WorldLocations, float]:

    if constraint is None:
        # this is a sort of "training wheels" condition, which is the only reason it comes before
        # the categoricals below
        is_path_direct, difficulty = select_boolean_difficulty(difficulty, rand)
    else:
        is_path_direct = False

    if is_path_direct:
        desired_goal_dist = scale_with_difficulty(
            difficulty, task_config.direct_goal_dist_easy, task_config.direct_goal_dist_hard
        )
        world, locations = create_world_from_constraint(
            stats.norm(desired_goal_dist, TIGHT_DIST_STD_DEV), rand, difficulty, export_config, constraint
        )
    else:
        path_point_count = round(
            normal_distrib_range(
                task_config.indirect_path_point_count_easy - 0.49,
                task_config.indirect_path_point_count_hard + 0.49,
                task_config.indirect_path_point_count_std_dev,
                rand,
                difficulty,
            )
        )
        desired_goal_dist = scale_with_difficulty(
            difficulty, task_config.indirect_goal_dist_easy, task_config.indirect_goal_dist_hard
        )
        traversal_width = normal_distrib_range(
            task_config.indirect_path_width_easy,
            task_config.indirect_path_width_hard,
            task_config.indirect_path_width_std_dev,
            rand,
            difficulty,
        )

        world, locations = create_world_from_constraint(
            stats.norm(desired_goal_dist, TIGHT_DIST_STD_DEV), rand, difficulty, export_config, constraint
        )

        max_gap_dist, _warnings = world.get_critical_distance(locations, _MIN_GAP_DISTANCE)

        if max_gap_dist is None:
            raise WorldTooSmall(AvalonTask.MOVE, _MIN_GAP_DISTANCE, locations.get_2d_spawn_goal_distance())

        if max_gap_dist is not None:
            gap_distance = normal_distrib_range(
                max_gap_dist * 0.3, max_gap_dist * 0.8, max_gap_dist * 0.1, rand, difficulty
            )
            depth = difficulty_variation(
                task_config.indirect_chasm_depth_easy, task_config.indirect_chasm_depth_hard, rand, difficulty
            )
            ring_config = make_ring(
                rand,
                difficulty,
                world,
                locations,
                height=-depth,
                gap_distance=gap_distance,
                traversal_width=traversal_width,
                dual_solution=HeightSolution(
                    paths=(
                        HeightPath(
                            is_path_restricted_to_land=True,
                            extra_point_count=path_point_count,
                            width=traversal_width,
                            is_path_climbable=True,
                            is_outside_edge_unclimbable=True,
                            # sometimes your paths will be slightly simpler than you expected.
                            # in those cases, a simplicity warning will be logged
                            is_path_failure_allowed=True,
                        ),
                    ),
                    solution_point_brink_distance=1.0,
                ),
                traversal_noise_interpolation_multiple=4.0,
                constraint=constraint,
            )

            world = world.add_height_obstacle(rand, ring_config, locations.island)

    return world, locations, difficulty



from typing import Tuple
from typing import cast

import numpy as np

from avalon.datagen.world_creation.types import MapFloatNP


# from: https://stackoverflow.com/questions/42147776/producing-2d-perlin-noise-with-numpy
def perlin(
    shape: Tuple[int, ...],
    scale: float,
    rand: np.random.Generator,
    is_normalized: bool = False,
    noise_min: float = 0.0,
) -> MapFloatNP:
    assert len(shape) == 2  # Ensures type-checking passes when passing `some_array.shape` directly
    cache_key = (scale,) + shape
    # create the grid
    y_lin = np.linspace(0, scale * shape[0], shape[0], endpoint=False)
    x_lin = np.linspace(0, scale * shape[1], shape[1], endpoint=False)
    x, y = np.meshgrid(x_lin, y_lin)

    # permutation table
    p = np.arange(max([round(scale * shape[0]), round(scale * shape[1])]) + 1, dtype=int)
    rand.shuffle(p)
    p = np.stack([p, p]).flatten()
    # coordinates of the top-left
    xi, yi = x.astype(int), y.astype(int)
    # internal coordinates
    xf, yf = x - xi, y - yi
    # fade factors
    u, v = fade(xf), fade(yf)
    # noise components
    n00 = gradient(p[p[xi] + yi], xf, yf)
    n01 = gradient(p[p[xi] + yi + 1], xf, yf - 1)
    n11 = gradient(p[p[xi + 1] + yi + 1], xf - 1, yf - 1)
    n10 = gradient(p[p[xi + 1] + yi], xf - 1, yf)
    # combine noises
    x1 = lerp(n00, n10, u)
    x2 = lerp(n01, n11, u)  # FIX1: I was using n10 instead of n01
    result = lerp(x1, x2, v)  # FIX2: I also had to reverse x1 and x2 here

    if is_normalized:
        result -= np.min(result)
        result /= np.max(result)

        if noise_min > 0.0:
            result /= 1.0 - noise_min
            result += noise_min

    return cast(MapFloatNP, result)


def lerp(a, b, x):  # type: ignore
    "linear interpolation"
    return a + x * (b - a)


def fade(t):  # type: ignore
    "6t^5 - 15t^4 + 10t^3"
    return 6 * t**5 - 15 * t**4 + 10 * t**3


def gradient(h, x, y):  # type: ignore
    "grad converts h to the right gradient vector and return the dot product with (x,y)"
    vectors = np.array([[0, 1], [0, -1], [1, 0], [-1, 0]])
    g = vectors[h % 4]
    return g[:, :, 0] * x + g[:, :, 1] * y



import attr
import numpy as np
from godot_parser import Node as GDNode

from avalon.common.utils import to_immutable_array
from avalon.datagen.world_creation.entities.tools.tool import Tool
from avalon.datagen.world_creation.types import GodotScene


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Placeholder(Tool):
    position: np.ndarray = attr.ib(
        default=np.array([0.0, 0.0, 0.0]),
        converter=to_immutable_array,
        eq=attr.cmp_using(eq=np.array_equal),  # type: ignore[attr-defined]
    )
    entity_id: int = -1
    resource_file: str = "res://items/not_real.tscn"
    offset: float = 0.0

    def get_offset(self) -> float:
        return self.offset

    def get_node(self, scene: GodotScene) -> GDNode:
        raise Exception("Not implemented for a reason--should never be called!")



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import logging
import os
import pprint
import threading
import time
import timeit
import traceback
import typing

import gym.spaces.utils
import numpy as np
import torch
from gym import spaces
from torch import multiprocessing as mp
from torch import nn
from torch.nn import functional as F

from avalon.agent.godot.godot_gym import AvalonEnv
from avalon.agent.torchbeast import atari_wrappers
from avalon.agent.torchbeast import avalon_helpers
from avalon.agent.torchbeast.atari_wrappers import WarpFrame
from avalon.agent.torchbeast.avalon_helpers import FlattenAction
from avalon.agent.torchbeast.avalon_helpers import GodotToPyTorch
from avalon.agent.torchbeast.avalon_helpers import compute_policy_gradient_loss_from_dist
from avalon.agent.torchbeast.avalon_helpers import flatten_tensor
from avalon.agent.torchbeast.avalon_helpers import unflatten_tensor
from avalon.agent.torchbeast.avalon_helpers import vtrace_from_dist
from avalon.agent.torchbeast.core import environment
from avalon.agent.torchbeast.core import file_writer
from avalon.agent.torchbeast.core import prof
from avalon.agent.torchbeast.model import DictActionHead
from avalon.common.log_utils import configure_parent_logging
from avalon.datagen.godot_env.godot_env import _randint_of_size

os.environ["OMP_NUM_THREADS"] = "1"  # Necessary for multithreading.


def get_flags():
    # yapf: disable
    parser = argparse.ArgumentParser(description="PyTorch Scalable Agent")

    parser.add_argument("--env", type=str, default="PongNoFrameskip-v4",
                        help="Gym environment.")
    parser.add_argument("--mode", default="train",
                        choices=["train", "test", "test_render"],
                        help="Training or test mode.")
    parser.add_argument("--xpid", default=None,
                        help="Experiment id (default: None).")

    # Training settings.
    parser.add_argument("--disable_checkpoint", action="store_true",
                        help="Disable saving checkpoint.")
    parser.add_argument("--savedir", default="~/logs/torchbeast",
                        help="Root dir where experiment data will be saved.")
    parser.add_argument("--num_actors", default=4, type=int, metavar="N",
                        help="Number of actors (default: 4).")
    parser.add_argument("--total_steps", default=100000, type=int, metavar="T",
                        help="Total environment steps to train for.")
    parser.add_argument("--batch_size", default=8, type=int, metavar="B",
                        help="Learner batch size.")
    parser.add_argument("--unroll_length", default=80, type=int, metavar="T",
                        help="The unroll length (time dimension).")
    parser.add_argument("--num_buffers", default=None, type=int,
                        metavar="N", help="Number of shared-memory buffers.")
    parser.add_argument("--num_learner_threads", "--num_threads", default=2, type=int,
                        metavar="N", help="Number learner threads.")
    parser.add_argument("--disable_cuda", action="store_true",
                        help="Disable CUDA.")
    parser.add_argument("--use_lstm", action="store_true",
                        help="Use LSTM in agent model.")

    # Loss settings.
    parser.add_argument("--entropy_cost", default=0.0006,
                        type=float, help="Entropy cost/multiplier.")
    parser.add_argument("--baseline_cost", default=0.5,
                        type=float, help="Baseline cost/multiplier.")
    parser.add_argument("--discounting", default=0.99,
                        type=float, help="Discounting factor.")
    parser.add_argument("--reward_clipping", default="abs_one",
                        choices=["abs_one", "none"],
                        help="Reward clipping.")

    # Optimizer settings.
    parser.add_argument("--learning_rate", default=0.00048,
                        type=float, metavar="LR", help="Learning rate.")
    parser.add_argument("--alpha", default=0.99, type=float,
                        help="RMSProp smoothing constant.")
    parser.add_argument("--momentum", default=0, type=float,
                        help="RMSProp momentum.")
    parser.add_argument("--epsilon", default=0.01, type=float,
                        help="RMSProp epsilon.")
    parser.add_argument("--grad_norm_clipping", default=40.0, type=float,
                        help="Global gradient norm clip.")
    # yapf: enable

    namespace = argparse.Namespace()
    flags, argv = parser.parse_known_args(namespace=namespace)
    return flags


logging.basicConfig(
    format=("[%(levelname)s:%(process)d %(module)s:%(lineno)d %(asctime)s] " "%(message)s"),
    level=0,
)

Buffers = typing.Dict[str, typing.List[torch.Tensor]]


def compute_baseline_loss(advantages):
    return 0.5 * torch.sum(advantages**2)


def compute_entropy_loss(logits):
    """Return the entropy loss, i.e., the negative entropy of the policy."""
    policy = F.softmax(logits, dim=-1)
    log_policy = F.log_softmax(logits, dim=-1)
    return torch.sum(policy * log_policy)


def compute_policy_gradient_loss(logits, actions, advantages):
    cross_entropy = F.nll_loss(
        F.log_softmax(torch.flatten(logits, 0, 1), dim=-1),
        target=torch.flatten(actions, 0, 1),
        reduction="none",
    )
    cross_entropy = cross_entropy.view_as(advantages)
    return torch.sum(cross_entropy * advantages.detach())


def act(
    flags,
    actor_index: int,
    free_queue: mp.SimpleQueue,
    full_queue: mp.SimpleQueue,
    model: torch.nn.Module,
    buffers: Buffers,
    initial_agent_state_buffers,
):
    try:
        configure_parent_logging()
        logging.info("Actor %i started.", actor_index)
        timings = prof.Timings()  # Keep track of how fast things are.

        gym_env = create_env(flags)
        seed = actor_index ^ int.from_bytes(os.urandom(4), byteorder="little")
        gym_env.seed(seed)
        env = environment.Environment(gym_env)
        env_output = env.initial()
        agent_state = model.initial_state(batch_size=1)
        agent_output, unused_state = model(env_output, agent_state)
        while True:
            index = free_queue.get()
            if index is None:
                break

            # Write old rollout end.
            for key in env_output:
                buffers[key][index][0, ...] = env_output[key]
            for key in agent_output:
                buffers[key][index][0, ...] = agent_output[key]
            for i, tensor in enumerate(agent_state):
                initial_agent_state_buffers[index][i][...] = tensor

            # Do new rollout.
            for t in range(flags.unroll_length):
                timings.reset()

                with torch.no_grad():
                    agent_output, agent_state = model(env_output, agent_state)

                timings.time("model")

                env_output = env.step(agent_output["action"])

                timings.time("step")

                for key in env_output:
                    # logging.debug(key)
                    # if key == "action":
                    #     logging.warning(f"action shapes: {buffers[key][index].shape}, {env_output[key].shape}")
                    buffers[key][index][t + 1, ...] = env_output[key]
                for key in agent_output:
                    buffers[key][index][t + 1, ...] = agent_output[key]

                timings.time("write")
            full_queue.put(index)

        if actor_index == 0:
            logging.info("Actor %i: %s", actor_index, timings.summary())

    except KeyboardInterrupt:
        pass  # Return silently.
    except Exception as e:
        logging.error("Exception in worker process %i", actor_index)
        traceback.print_exc()
        raise e


def get_batch(
    flags,
    free_queue: mp.SimpleQueue,
    full_queue: mp.SimpleQueue,
    buffers: Buffers,
    initial_agent_state_buffers,
    timings,
    lock=threading.Lock(),
):
    with lock:
        timings.time("lock")
        indices = [full_queue.get() for _ in range(flags.batch_size)]
        timings.time("dequeue")
    batch = {key: torch.stack([buffers[key][m] for m in indices], dim=1) for key in buffers}
    initial_agent_state = (torch.cat(ts, dim=1) for ts in zip(*[initial_agent_state_buffers[m] for m in indices]))
    timings.time("batch")
    for m in indices:
        free_queue.put(m)
    timings.time("enqueue")
    batch = {k: t.to(device=flags.device, non_blocking=True) for k, t in batch.items()}
    initial_agent_state = tuple(t.to(device=flags.device, non_blocking=True) for t in initial_agent_state)
    timings.time("device")
    return batch, initial_agent_state


def learn(
    flags,
    actor_model,
    model,
    batch,
    initial_agent_state,
    optimizer,
    scheduler,
    lock=threading.Lock(),  # noqa: B008
):
    """Performs a learning (optimization) step."""
    with lock:
        learner_outputs, unused_state = model(batch, initial_agent_state)

        # Take final value function slice for bootstrapping.
        bootstrap_value = learner_outputs["baseline"][-1]

        # Move from obs[t] -> action[t] to action[t] -> obs[t].
        batch = {key: tensor[1:] for key, tensor in batch.items()}
        learner_outputs = {key: tensor[:-1] for key, tensor in learner_outputs.items()}

        rewards = batch["reward"]
        if flags.reward_clipping == "abs_one":
            clipped_rewards = torch.clamp(rewards, -1, 1)
        elif flags.reward_clipping == "none":
            clipped_rewards = rewards

        discounts = (~batch["done"]).float() * flags.discounting

        # TODO: explode if action head has learnable parameters
        behavior_policy_dist = model.action_head(batch["policy_logits"])
        target_policy_dist = model.action_head(learner_outputs["policy_logits"])
        action_dict = model.unflatten(batch["action"])
        vtrace_returns = vtrace_from_dist(
            behavior_policy_dist=behavior_policy_dist,
            target_policy_dist=target_policy_dist,
            actions=action_dict,
            discounts=discounts,
            rewards=clipped_rewards,
            values=learner_outputs["baseline"],
            bootstrap_value=bootstrap_value,
        )

        pg_loss = compute_policy_gradient_loss_from_dist(
            target_policy_dist,
            action_dict,
            vtrace_returns.pg_advantages,
        )
        baseline_loss = flags.baseline_cost * compute_baseline_loss(vtrace_returns.vs - learner_outputs["baseline"])
        entropy_loss = flags.entropy_cost * compute_entropy_loss(learner_outputs["policy_logits"])

        total_loss = pg_loss + baseline_loss + entropy_loss

        episode_returns = batch["episode_return"][batch["done"]]
        stats = {
            "episode_returns": tuple(episode_returns.cpu().numpy()),
            "mean_episode_return": torch.mean(episode_returns).item(),
            "total_loss": total_loss.item(),
            "pg_loss": pg_loss.item(),
            "baseline_loss": baseline_loss.item(),
            "entropy_loss": entropy_loss.item(),
        }

        optimizer.zero_grad()
        total_loss.backward()
        nn.utils.clip_grad_norm_(model.parameters(), flags.grad_norm_clipping)
        optimizer.step()
        scheduler.step()

        actor_model.load_state_dict(model.state_dict())
        return stats


def create_buffers(flags, obs_shape, num_policy_logits, flat_action_dim) -> Buffers:
    T = flags.unroll_length

    specs = dict(
        frame=dict(size=(T + 1, *obs_shape), dtype=torch.uint8),
        reward=dict(size=(T + 1,), dtype=torch.float32),
        done=dict(size=(T + 1,), dtype=torch.bool),
        episode_return=dict(size=(T + 1,), dtype=torch.float32),
        episode_step=dict(size=(T + 1,), dtype=torch.int32),
        policy_logits=dict(size=(T + 1, num_policy_logits), dtype=torch.float32),
        baseline=dict(size=(T + 1,), dtype=torch.float32),
        last_action=dict(size=(T + 1, flat_action_dim), dtype=torch.float32),
        action=dict(size=(T + 1, flat_action_dim), dtype=torch.float32),
    )
    buffers: Buffers = {key: [] for key in specs}
    for _ in range(flags.num_buffers):
        for key in buffers:
            buffers[key].append(torch.empty(**specs[key]).share_memory_())
    return buffers


def train(flags):  # pylint: disable=too-many-branches, too-many-statements
    if flags.xpid is None:
        flags.xpid = "torchbeast-%s" % time.strftime("%Y%m%d-%H%M%S")
    plogger = file_writer.FileWriter(xpid=flags.xpid, xp_args=flags.__dict__, rootdir=flags.savedir)
    checkpointpath = os.path.expandvars(os.path.expanduser("%s/%s/%s" % (flags.savedir, flags.xpid, "model.tar")))

    if flags.num_buffers is None:  # Set sensible default for num_buffers.
        flags.num_buffers = max(2 * flags.num_actors, flags.batch_size)
    if flags.num_actors >= flags.num_buffers:
        raise ValueError("num_buffers should be larger than num_actors")
    if flags.num_buffers < flags.batch_size:
        raise ValueError("num_buffers should be larger than batch_size")

    T = flags.unroll_length
    B = flags.batch_size

    flags.device = None
    if not flags.disable_cuda and torch.cuda.is_available():
        logging.info("Using CUDA.")
        flags.device = torch.device("cuda")
    else:
        logging.info("Not using CUDA.")
        flags.device = torch.device("cpu")

    env = create_env(flags)

    model = Net(env.observation_space.shape, env.action_space, flags.use_lstm)

    buffers = create_buffers(
        flags, env.observation_space.shape, model.num_actions, gym.spaces.utils.flatdim(env.action_space)
    )

    model.share_memory()

    # Add initial RNN state.
    initial_agent_state_buffers = []
    for _ in range(flags.num_buffers):
        state = model.initial_state(batch_size=1)
        for t in state:
            t.share_memory_()
        initial_agent_state_buffers.append(state)

    actor_processes = []
    ctx = mp.get_context("fork")
    free_queue = ctx.SimpleQueue()
    full_queue = ctx.SimpleQueue()

    for i in range(flags.num_actors):
        actor = ctx.Process(
            target=act,
            args=(
                flags,
                i,
                free_queue,
                full_queue,
                model,
                buffers,
                initial_agent_state_buffers,
            ),
        )
        actor.start()
        actor_processes.append(actor)

    learner_model = Net(env.observation_space.shape, env.action_space, flags.use_lstm).to(device=flags.device)

    optimizer = torch.optim.RMSprop(
        learner_model.parameters(),
        lr=flags.learning_rate,
        momentum=flags.momentum,
        eps=flags.epsilon,
        alpha=flags.alpha,
    )

    def lr_lambda(epoch):
        return 1 - min(epoch * T * B, flags.total_steps) / flags.total_steps

    scheduler = torch.optim.lr_scheduler.LambdaLR(optimizer, lr_lambda)

    logger = logging.getLogger("logfile")
    stat_keys = [
        "total_loss",
        "mean_episode_return",
        "pg_loss",
        "baseline_loss",
        "entropy_loss",
    ]
    logger.info("# Step\t%s", "\t".join(stat_keys))

    step, stats = 0, {}

    def batch_and_learn(i, lock=threading.Lock()) -> None:
        """Thread target for the learning process."""
        nonlocal step, stats
        timings = prof.Timings()
        while step < flags.total_steps:
            timings.reset()
            batch, agent_state = get_batch(
                flags,
                free_queue,
                full_queue,
                buffers,
                initial_agent_state_buffers,
                timings,
            )
            stats = learn(
                flags,
                model,
                learner_model,
                batch,
                agent_state,
                optimizer,
                scheduler,
            )
            timings.time("learn")
            with lock:
                to_log = dict(step=step)
                to_log.update({k: stats[k] for k in stat_keys})
                plogger.log(to_log)
                step += T * B

        if i == 0:
            logging.info("Batch and learn: %s", timings.summary())

    for m in range(flags.num_buffers):
        free_queue.put(m)

    threads = []
    for i in range(flags.num_learner_threads):
        thread = threading.Thread(target=batch_and_learn, name="batch-and-learn-%d" % i, args=(i,))
        thread.start()
        threads.append(thread)

    def checkpoint() -> None:
        if flags.disable_checkpoint:
            return
        logging.info("Saving checkpoint to %s", checkpointpath)
        torch.save(
            {
                "model_state_dict": model.state_dict(),
                "optimizer_state_dict": optimizer.state_dict(),
                "scheduler_state_dict": scheduler.state_dict(),
                "flags": vars(flags),
            },
            checkpointpath,
        )

    timer = timeit.default_timer
    try:
        last_checkpoint_time = timer()
        while step < flags.total_steps:
            start_step = step
            start_time = timer()
            time.sleep(5)

            if timer() - last_checkpoint_time > 10 * 60:  # Save every 10 min.
                checkpoint()
                last_checkpoint_time = timer()

            sps = (step - start_step) / (timer() - start_time)
            if stats.get("episode_returns", None):
                mean_return = "Return per episode: %.1f. " % stats["mean_episode_return"]
            else:
                mean_return = ""
            total_loss = stats.get("total_loss", float("inf"))
            logging.info(
                "Steps %i @ %.1f SPS. Loss %f. %sStats:\n%s",
                step,
                sps,
                total_loss,
                mean_return,
                pprint.pformat(stats),
            )
    except KeyboardInterrupt:
        return  # Try joining actors then quit.
    else:
        for thread in threads:
            thread.join()
        logging.info("Learning finished after %d steps.", step)
    finally:
        for _ in range(flags.num_actors):
            free_queue.put(None)
        for actor in actor_processes:
            actor.join(timeout=1)

    checkpoint()
    plogger.close()


def test(flags, num_episodes: int = 10) -> None:
    if flags.xpid is None:
        checkpointpath = "./latest/model.tar"
    else:
        checkpointpath = os.path.expandvars(os.path.expanduser("%s/%s/%s" % (flags.savedir, flags.xpid, "model.tar")))

    gym_env = create_env(flags)
    env = environment.Environment(gym_env)
    model = Net(gym_env.observation_space.shape, gym_env.action_space, flags.use_lstm)
    model.eval()
    checkpoint = torch.load(checkpointpath, map_location="cpu")
    model.load_state_dict(checkpoint["model_state_dict"])

    observation = env.initial()
    returns = []

    while len(returns) < num_episodes:
        if flags.mode == "test_render":
            env.gym_env.render()
        agent_outputs = model(observation)
        policy_outputs, _ = agent_outputs
        observation = env.step(policy_outputs["action"])
        if observation["done"].item():
            returns.append(observation["episode_return"].item())
            logging.info(
                "Episode ended after %d steps. Return: %.1f",
                observation["episode_step"].item(),
                observation["episode_return"].item(),
            )
    env.close()
    logging.info("Average returns over %i steps: %.1f", num_episodes, sum(returns) / len(returns))


class AtariNet(nn.Module):
    def __init__(self, observation_shape, action_space: spaces.Space, use_lstm=False) -> None:
        super(AtariNet, self).__init__()
        self.observation_shape = observation_shape

        self._action_space = action_space
        self.action_head = DictActionHead(action_space)
        num_actions = self.action_head.num_inputs
        self.num_actions = num_actions
        self.last_action_dim = spaces.flatdim(action_space)

        # Feature extraction.
        self.conv1 = nn.Conv2d(
            in_channels=self.observation_shape[0],
            out_channels=32,
            kernel_size=8,
            stride=4,
        )
        self.conv2 = nn.Conv2d(32, 64, kernel_size=4, stride=2)
        self.conv3 = nn.Conv2d(64, 64, kernel_size=3, stride=1)

        # Fully connected layer.
        self.fc = nn.Linear(3136, 512)

        # FC output size + one-hot of last action + last reward.
        core_output_size = self.fc.out_features + self.last_action_dim + 1

        self.use_lstm = use_lstm
        if use_lstm:
            self.core = nn.LSTM(core_output_size, core_output_size, 2)

        self.policy = nn.Linear(core_output_size, self.num_actions)
        self.baseline = nn.Linear(core_output_size, 1)

    def initial_state(self, batch_size):
        if not self.use_lstm:
            return tuple()
        return tuple(torch.zeros(self.core.num_layers, batch_size, self.core.hidden_size) for _ in range(2))

    def forward(self, inputs, core_state=()):
        x = inputs["frame"]  # [T, B, C, H, W].
        T, B, *_ = x.shape
        x = torch.flatten(x, 0, 1)  # Merge time and batch.
        x = x.float() / 255.0
        x = F.relu(self.conv1(x))
        x = F.relu(self.conv2(x))
        x = F.relu(self.conv3(x))
        x = x.view(T * B, -1)
        x = F.relu(self.fc(x))

        # one_hot_last_action = F.one_hot(inputs["last_action"].view(T * B), self.num_actions).float()
        one_hot_last_action = inputs["last_action"].view(T * B, self.last_action_dim)
        clipped_reward = torch.clamp(inputs["reward"], -1, 1).view(T * B, 1)
        core_input = torch.cat([x, clipped_reward, one_hot_last_action], dim=-1)

        if self.use_lstm:
            core_input = core_input.view(T, B, -1)
            core_output_list = []
            notdone = (~inputs["done"]).float()
            for input, nd in zip(core_input.unbind(), notdone.unbind()):
                # Reset core state to zero whenever an episode ended.
                # Make `done` broadcastable with (num_layers, B, hidden_size)
                # states:
                nd = nd.view(1, -1, 1)
                core_state = tuple(nd * s for s in core_state)
                output, core_state = self.core(input.unsqueeze(0), core_state)
                core_output_list.append(output)
            core_output = torch.flatten(torch.cat(core_output_list), 0, 1)
        else:
            core_output = core_input
            core_state = tuple()

        policy_logits = self.policy(core_output)
        baseline = self.baseline(core_output)

        action = self.action_head(policy_logits).sample()

        # if self.training:
        #     action = torch.multinomial(F.softmax(policy_logits, dim=1), num_samples=1)
        # else:
        #     # Don't sample when testing.
        #     action = torch.argmax(policy_logits, dim=1)

        policy_logits = policy_logits.view(T, B, self.num_actions)
        baseline = baseline.view(T, B)
        # action = action.view(T, B)
        flat_action = self.flatten(action)

        return (
            dict(policy_logits=policy_logits, baseline=baseline, action=flat_action),
            core_state,
        )

    def flatten(self, action):
        return flatten_tensor(self._action_space, action)

    def unflatten(self, flat_action):
        return unflatten_tensor(self._action_space, flat_action)


Net = AtariNet


def create_env(flags):
    if flags.env == "godot":
        seed = _randint_of_size(np.uint32)
        config = dict(random_int=seed, time_limit=100)
        env = AvalonEnv(config)
        env = WarpFrame(env, grayscale=False, dict_space_key="rgb")
        env = FlattenAction(env)
        env = GodotToPyTorch(env)
        # env = CurriculumWrapper(env, update_rate=args.difficulty_update_rate, std=args.difficulty_std)
        return env
    else:
        return atari_wrappers.wrap_pytorch(
            avalon_helpers.wrap_atari(
                atari_wrappers.wrap_deepmind(
                    atari_wrappers.make_atari(flags.env),
                    clip_rewards=False,
                    frame_stack=True,
                    scale=False,
                )
            )
        )


def main(flags) -> None:
    if flags.mode == "train":
        train(flags)
    else:
        test(flags)


if __name__ == "__main__":
    flags = get_flags()
    main(flags)



import math
from enum import Enum
from itertools import product
from pathlib import Path
from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Tuple

import attr
import numpy as np
from scipy import stats

from avalon.common.errors import SwitchError
from avalon.common.log_utils import logger
from avalon.common.utils import only
from avalon.datagen.errors import ImpossibleWorldError
from avalon.datagen.errors import WorldTooSmall
from avalon.datagen.world_creation.configs.building import BuildingConfig
from avalon.datagen.world_creation.configs.export import ExportConfig
from avalon.datagen.world_creation.configs.task import IndoorTaskConfig
from avalon.datagen.world_creation.configs.task import TaskConfig
from avalon.datagen.world_creation.constants import AGENT_HEIGHT
from avalon.datagen.world_creation.constants import BOULDER_HEIGHT
from avalon.datagen.world_creation.constants import BOULDER_MAX_MASS
from avalon.datagen.world_creation.constants import BOULDER_MIN_MASS
from avalon.datagen.world_creation.constants import FOOD_HOVER_DIST
from avalon.datagen.world_creation.constants import JUMPING_REQUIRED_HEIGHT
from avalon.datagen.world_creation.constants import MAX_FLAT_JUMP_DIST
from avalon.datagen.world_creation.constants import MAX_FLAT_JUMP_METERS
from avalon.datagen.world_creation.constants import MAX_JUMP_HEIGHT_METERS
from avalon.datagen.world_creation.constants import MIN_BRIDGE_DIST
from avalon.datagen.world_creation.constants import AvalonTask
from avalon.datagen.world_creation.entities.constants import CANONICAL_FOOD_HEIGHT_ON_TREE
from avalon.datagen.world_creation.entities.tools.boulder import Boulder
from avalon.datagen.world_creation.geometry import BuildingTile
from avalon.datagen.world_creation.geometry import euclidean_distance
from avalon.datagen.world_creation.geometry import local_to_global_coords
from avalon.datagen.world_creation.geometry import midpoint
from avalon.datagen.world_creation.geometry import squares_overlap
from avalon.datagen.world_creation.indoor.builders import DefaultHallwayBuilder
from avalon.datagen.world_creation.indoor.builders import HouseLikeRoomBuilder
from avalon.datagen.world_creation.indoor.builders import RectangleFootprintBuilder
from avalon.datagen.world_creation.indoor.builders import WindowBuilder
from avalon.datagen.world_creation.indoor.builders import get_free_obstacle_sites
from avalon.datagen.world_creation.indoor.building import Building
from avalon.datagen.world_creation.indoor.building import BuildingAestheticsConfig
from avalon.datagen.world_creation.indoor.building import BuildingTask
from avalon.datagen.world_creation.indoor.components import Room
from avalon.datagen.world_creation.indoor.components import Wall
from avalon.datagen.world_creation.indoor.components import Window
from avalon.datagen.world_creation.indoor.constants import CEILING_THICKNESS
from avalon.datagen.world_creation.indoor.constants import DEFAULT_FLOOR_THICKNESS
from avalon.datagen.world_creation.indoor.constants import Azimuth
from avalon.datagen.world_creation.indoor.task_generator import CANONICAL_BUILDING_LOCATION
from avalon.datagen.world_creation.indoor.task_generator import BuildingTaskGenerator
from avalon.datagen.world_creation.indoor.task_generator import IndoorTaskParams
from avalon.datagen.world_creation.indoor.task_generator import _decide_boulder_mass
from avalon.datagen.world_creation.indoor.task_generator import add_food_island
from avalon.datagen.world_creation.indoor.task_generator import create_building_for_skill_scenario
from avalon.datagen.world_creation.indoor.task_generator import decide_spawn_room_and_tile
from avalon.datagen.world_creation.indoor.task_generator import make_indoor_task_world
from avalon.datagen.world_creation.indoor.task_generator import rectangle_dimensions_within_radius
from avalon.datagen.world_creation.indoor.tiles import decide_tiles_by_distance
from avalon.datagen.world_creation.indoor.tiles import draw_line_in_grid
from avalon.datagen.world_creation.indoor.tiles import tile_centroid
from avalon.datagen.world_creation.tasks.eat import add_food_tree_for_simple_task
from avalon.datagen.world_creation.types import HeightMode
from avalon.datagen.world_creation.utils import decompose_weighted_mean
from avalon.datagen.world_creation.utils import to_2d_point
from avalon.datagen.world_creation.worlds.creation import create_world_from_constraint
from avalon.datagen.world_creation.worlds.difficulty import normal_distrib_range
from avalon.datagen.world_creation.worlds.difficulty import scale_with_difficulty
from avalon.datagen.world_creation.worlds.difficulty import select_boolean_difficulty
from avalon.datagen.world_creation.worlds.export import export_world
from avalon.datagen.world_creation.worlds.obstacles.configure import make_ring
from avalon.datagen.world_creation.worlds.obstacles.height_path import HeightPath
from avalon.datagen.world_creation.worlds.obstacles.height_solution import HeightSolution
from avalon.datagen.world_creation.worlds.types import CompositionalConstraint
from avalon.datagen.world_creation.worlds.utils import add_offsets
from avalon.datagen.world_creation.worlds.world import World
from avalon.datagen.world_creation.worlds.world_locations import WorldLocations

_EXTRA_SAFETY_RADIUS = 3.0
_MIN_PUSH_TASK_DISTANCE = MIN_BRIDGE_DIST + 2 * _EXTRA_SAFETY_RADIUS


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class PushIndoorTaskConfig(IndoorTaskConfig):
    # Smallest and largest possible site radius for a building with this task to be placed on.
    min_site_radius: float = 10.0
    max_site_radius: float = 15.0
    # Standard deviation for the distribution deciding the site radius: higher means more variability at same difficulty
    site_radius_std_dev: float = 1.0
    # Total number of stories for the building. Note that this task is a single-story task, all other stories will
    # be purely aesthetic (e.g. to help with being viewed in the distance outdoors).
    story_count: int = 2
    # Boulder size that needs to be pushed; mainly determines visibility while pushing
    blocked_entrance_boulder_size: float = 2
    # Square island side length; should be large enough to prevent knocking food off by reaching or jumping
    climb_higher_food_island_size: int = 5


@attr.s(auto_attribs=True, collect_by_mro=True, hash=True)
class PushOutdoorTaskConfig(TaskConfig):
    # how large of a gap to use for the push task in difficulty 0.0 and 1.0 worlds respectively
    gap_dist_easy: float = _MIN_PUSH_TASK_DISTANCE
    gap_dist_hard: float = _MIN_PUSH_TASK_DISTANCE + MAX_FLAT_JUMP_DIST
    # how high to make the chasm. Unfortunately needs to be really quite a narrow range because you need to be able to
    # push the boulder in, then jump on it, and still have enough space to be able to jump out on the other side
    chasm_depth_easy: float = JUMPING_REQUIRED_HEIGHT
    chasm_depth_hard: float = BOULDER_HEIGHT + JUMPING_REQUIRED_HEIGHT - 0.5
    # how far away to possibly put the boulder
    boulder_dist_easy: float = 0.0
    boulder_dist_hard: float = 10.0
    # how heavy to make the boulder
    boulder_mass_easy: float = BOULDER_MIN_MASS
    boulder_mass_hard: float = BOULDER_MAX_MASS
    boulder_mass_std_dev: float = 10.0
    # the minimum chance (ie, at difficulty=1.0) that the inside wall will be climbable
    final_is_inside_climbable_probability: float = 0.2


@attr.s(auto_attribs=True, collect_by_mro=True, hash=True)
class PushTaskConfig(TaskConfig):
    # likelihood that the indoor variant of the task will be used
    indoor_probability: float = 0.2
    # the configs for each variant of the task
    indoor_config: PushIndoorTaskConfig = PushIndoorTaskConfig()
    outdoor_config: PushOutdoorTaskConfig = PushOutdoorTaskConfig()


def generate_push_task(
    rand: np.random.Generator,
    difficulty: float,
    output_path: Path,
    export_config: ExportConfig,
    task_config: PushTaskConfig = PushTaskConfig(),
) -> None:
    is_indoor = rand.uniform() < task_config.indoor_probability
    if is_indoor:
        building, entities, spawn_location, target_location = create_building_for_skill_scenario(
            rand,
            difficulty,
            PushTaskGenerator(task_config.indoor_config),
            position=CANONICAL_BUILDING_LOCATION,
            is_indoor_only=True,
        )
        world = make_indoor_task_world(
            building, entities, difficulty, spawn_location, target_location, rand, export_config
        )
    else:
        world, locations, difficulty = create_push_obstacle(
            rand, difficulty, export_config, task_config=task_config.outdoor_config
        )
        world, locations = world.end_height_obstacles(
            locations, is_accessible_from_water=False, is_spawn_region_climbable=False
        )
        world = add_food_tree_for_simple_task(world, locations)
        world = world.add_spawn(rand, difficulty, locations.spawn, locations.goal)

    export_world(output_path, rand, world)


def create_push_obstacle(
    rand: np.random.Generator,
    difficulty: float,
    export_config: ExportConfig,
    constraint: Optional[CompositionalConstraint] = None,
    task_config: PushOutdoorTaskConfig = PushOutdoorTaskConfig(),
) -> Tuple[World, WorldLocations, float]:

    is_inside_climbable, difficulty = select_boolean_difficulty(
        difficulty, rand, final_prob=task_config.final_is_inside_climbable_probability
    )

    desired_goal_dist = (task_config.gap_dist_easy * 3.0) + (
        task_config.gap_dist_easy * 4.0 * difficulty * rand.uniform()
    )

    world, locations = create_world_from_constraint(
        stats.norm(desired_goal_dist, 0.5), rand, difficulty, export_config, constraint
    )

    desired_gap_distance = scale_with_difficulty(difficulty, task_config.gap_dist_easy, task_config.gap_dist_hard)
    gap_distance, _warnings = world.get_critical_distance(locations, task_config.gap_dist_easy, desired_gap_distance)

    height = scale_with_difficulty(difficulty, task_config.chasm_depth_easy, task_config.chasm_depth_hard)

    if gap_distance is None:
        raise WorldTooSmall(AvalonTask.PUSH, task_config.gap_dist_easy, locations.get_2d_spawn_goal_distance())
    gap_distance -= 2 * _EXTRA_SAFETY_RADIUS
    if gap_distance < MIN_BRIDGE_DIST:
        raise WorldTooSmall(AvalonTask.PUSH, gap_distance, MIN_BRIDGE_DIST)
    logger.trace(f"Creating a {gap_distance} meter gap")
    randomization_dist = scale_with_difficulty(
        difficulty, task_config.boulder_dist_easy, task_config.boulder_dist_hard
    )
    boulder_mass = normal_distrib_range(
        task_config.boulder_mass_easy,
        task_config.boulder_mass_hard,
        task_config.boulder_mass_std_dev,
        rand,
        difficulty,
    )
    boulder = Boulder(position=np.array([-1.0, 0.0, 0.0]), mass=boulder_mass)
    ring_config = make_ring(
        rand,
        difficulty,
        world,
        locations,
        gap_distance,
        constraint=constraint,
        height=-height,
        traversal_width=normal_distrib_range(10.0, 3.0, 1.0, rand, difficulty),
        is_inside_climbable=True,
        is_outside_climbable=False,
        dual_solution=HeightSolution(
            solution_point_brink_distance=1.0,
            inside_items=tuple(add_offsets([boulder])),
            inside_item_randomization_distance=randomization_dist,
            inside_item_radius=1.0,
            paths=(
                HeightPath(
                    is_solution_flattened=True,
                    is_path_restricted_to_land=False,
                    is_chasm_bottom_flattened=True,
                    is_height_affected=False,
                    width=randomization_dist * 2.0,
                    flattening_mode="min",
                ),
                HeightPath(
                    is_solution_flattened=False,
                    is_path_restricted_to_land=True,
                    is_chasm_bottom_flattened=False,
                    is_height_affected=False,
                    width=2.0,
                ),
            ),
        ),
        probability_of_centering_on_spawn=0.0 if constraint is None else None,
        height_mode=HeightMode.MIDPOINT_ABSOLUTE,
        expansion_meters=_EXTRA_SAFETY_RADIUS,
    )
    world = world.add_height_obstacle(rand, ring_config, locations.island)
    new_goal = locations.goal.copy()
    new_goal[1] = world.get_height_at(to_2d_point(new_goal)) + CANONICAL_FOOD_HEIGHT_ON_TREE
    locations = attr.evolve(locations, goal=new_goal)
    return world, locations, difficulty


class PushTaskVariant(Enum):
    BLOCKED_ENTRANCE = "blocked_entrance"
    CLIMB_HIGHER = "climb_higher"
    CROSS_CHASM = "cross_chasm"


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class PushTaskGenerator(BuildingTaskGenerator):
    config: PushIndoorTaskConfig = PushIndoorTaskConfig()

    def get_site_radius(self, rand: np.random.Generator, difficulty: float) -> float:
        return normal_distrib_range(
            self.config.min_site_radius, self.config.max_site_radius, self.config.site_radius_std_dev, rand, difficulty
        )

    def get_building_config(
        self,
        rand: np.random.Generator,
        difficulty: float,
        radius: float,
        allowed_auxiliary_tasks: Tuple[BuildingTask, ...] = tuple(),
        aesthetics: BuildingAestheticsConfig = BuildingAestheticsConfig(),
    ) -> BuildingConfig:
        width, length = rectangle_dimensions_within_radius(radius)
        return BuildingConfig(
            width=width,
            length=length,
            story_count=self.config.story_count,
            footprint_builder=RectangleFootprintBuilder(),
            room_builder=HouseLikeRoomBuilder(max_rooms=1),
            hallway_builder=DefaultHallwayBuilder(),
            window_builder=WindowBuilder(),
            aesthetics=aesthetics,
        )

    def get_principal_obstacle_params(self, rand: np.random.Generator, difficulty: float, building: Building) -> Tuple:
        difficulty_weights = np.array([2, 1])
        variant_difficulty, param_difficulty = decompose_weighted_mean(difficulty, difficulty_weights, rand=rand)
        variants_by_difficulty = {
            0.25: PushTaskVariant.BLOCKED_ENTRANCE,
            0.5: PushTaskVariant.CLIMB_HIGHER,
            0.75: PushTaskVariant.CROSS_CHASM,
        }
        variant_difficulty_distribution = stats.norm(variant_difficulty, 0.125)
        difficulty_weights = np.array([variant_difficulty_distribution.pdf(x) for x in variants_by_difficulty.keys()])
        difficulty_weights /= difficulty_weights.sum()
        task_variant = rand.choice(list(variants_by_difficulty.values()), p=difficulty_weights)  # type: ignore[arg-type]

        build_fn: Callable[..., Any]
        boulder_mass = _decide_boulder_mass(rand, difficulty)
        if task_variant == PushTaskVariant.BLOCKED_ENTRANCE:
            build_fn = self._add_obstacles_blocked_entrance
            extra_build_params = self._get_blocked_entrance_build_params(building, param_difficulty, rand)
        elif task_variant == PushTaskVariant.CLIMB_HIGHER:
            build_fn = self._add_obstacles_climb_higher
            extra_build_params = self._get_climb_higher_build_params(building, param_difficulty, rand)
        elif task_variant == PushTaskVariant.CROSS_CHASM:
            build_fn = self._add_obstacles_cross_chasm
            extra_build_params = self._get_cross_chasm_build_params(building, param_difficulty, rand)
        else:
            raise SwitchError(task_variant)

        entrance_azimuth = Azimuth.EAST
        return build_fn, extra_build_params, boulder_mass, entrance_azimuth

    def _get_blocked_entrance_build_params(
        self, building: Building, difficulty: float, rand: np.random.Generator
    ) -> Dict:
        tile_size = 1
        boulder_size = self.config.blocked_entrance_boulder_size
        min_enclosure_width = math.ceil(boulder_size) + tile_size  # >1 extra tile to be able to walk around it

        first_story = building.stories[0]
        room = only(first_story.rooms)

        free_sites = [
            site
            for site in get_free_obstacle_sites(building.stories[:1])
            if min_enclosure_width <= site.site_index <= room.width - min_enclosure_width - 1 and site.vertical
        ]
        if len(free_sites) == 0:
            raise ImpossibleWorldError("Room too small to create task")
        site_idx = rand.choice(range(len(free_sites)))
        site = free_sites[site_idx]
        hole_tile = site.site_index, site.length // 2
        boulder_tile = hole_tile[0] + 1, hole_tile[1]

        # Spawn region is always on the east, so we can add a positive-x aligned entrance there easily
        # Target region excludes edges when possible to avoid grabbing food through walls
        target_region_x_range = range(site.site_index - 1)
        if len(target_region_x_range) > 1:
            target_region_x_range = range(1, site.site_index - 1)
        target_region_y_range = range(room.length - 1)
        if len(target_region_y_range) > 1:
            target_region_y_range = range(1, room.length - 1)
        target_region_tiles = list(product(target_region_x_range, target_region_y_range))
        spawn_region_tiles = list(product(range(site.site_index + 1, room.width), range(room.length)))
        free_spawn_region_tiles = [
            tile
            for tile in spawn_region_tiles
            if not squares_overlap(
                tile_centroid(tile), tile_size, tile_centroid(boulder_tile, boulder_size), boulder_size
            )
        ]
        spawn_room, spawn_tile = decide_spawn_room_and_tile(rand, first_story, {room.id: free_spawn_region_tiles})
        target_tile = only(decide_tiles_by_distance(target_region_tiles, spawn_tile, difficulty, rand))

        extra_build_params = dict(
            hole_tile=hole_tile,
            spawn_tile=spawn_tile,
            boulder_size=boulder_size,
            target_tile=target_tile,
        )
        return extra_build_params

    def _get_climb_higher_build_params(self, building: Building, difficulty: float, rand: np.random.Generator) -> Dict:
        first_story = building.stories[0]
        spawn_room, spawn_tile = decide_spawn_room_and_tile(rand, first_story)

        food_island_size = self.config.climb_higher_food_island_size
        viable_food_island_tiles = list(
            product(range(spawn_room.width - food_island_size - 1), range(spawn_room.length - food_island_size - 1))
        )
        if len(viable_food_island_tiles) == 0:
            raise ImpossibleWorldError("Building too small for push task")
        food_island_top_left_tile = only(
            decide_tiles_by_distance(viable_food_island_tiles, spawn_tile, difficulty, rand)
        )
        food_island_position = BuildingTile(*food_island_top_left_tile)
        food_island_tiles = list(
            product(
                range(food_island_position.x, food_island_position.x + food_island_size),
                range(food_island_position.z, food_island_position.z + food_island_size),
            )
        )

        leeway_factor = 0.95
        boulder_size = MAX_JUMP_HEIGHT_METERS * leeway_factor
        boulder_size_in_tiles = math.ceil(boulder_size)
        food_island_height = boulder_size + MAX_JUMP_HEIGHT_METERS * leeway_factor
        viable_boulder_tiles = [
            tile
            for tile in list(
                product(
                    range(boulder_size_in_tiles, spawn_room.width - boulder_size_in_tiles + 1),
                    range(boulder_size_in_tiles, spawn_room.length - boulder_size_in_tiles + 1),
                )
            )
            if tile not in food_island_tiles and tile != spawn_tile
        ]
        food_island_center_tile = (
            food_island_position.x + food_island_size // 2,
            food_island_position.z + food_island_size // 2,
        )
        if len(viable_boulder_tiles) == 0:
            raise ImpossibleWorldError("Building too small for push task")
        boulder_tile = only(decide_tiles_by_distance(viable_boulder_tiles, food_island_center_tile, difficulty, rand))

        extra_build_params = dict(
            spawn_tile=spawn_tile,
            food_island_top_left_tile=food_island_top_left_tile,
            food_island_size=food_island_size,
            food_island_height=food_island_height,
            boulder_tile=boulder_tile,
            boulder_size=boulder_size,
        )
        return extra_build_params

    def _get_cross_chasm_build_params(self, building: Building, difficulty: float, rand: np.random.Generator) -> Dict:
        boulder_size = MAX_JUMP_HEIGHT_METERS * 0.9
        chasm_depth = (MAX_JUMP_HEIGHT_METERS * 1.8) - DEFAULT_FLOOR_THICKNESS
        chasm_size = round(MAX_FLAT_JUMP_METERS * 2)
        min_launch_side_distance_from_wall = math.ceil(boulder_size)
        min_landing_side_distance_from_wall = 3  # at least 1 tile away from wall and at least 1 tile away from chasm

        tile_size = 1
        first_story = building.stories[0]
        room = only(first_story.rooms)

        viable_x_positions = list(
            range(min_landing_side_distance_from_wall, room.width - min_launch_side_distance_from_wall - chasm_size)
        )
        if len(viable_x_positions) == 0:
            raise ImpossibleWorldError("Room too small to place a chasm")
        chasm_left_x = rand.choice(viable_x_positions)

        spawn_region_tiles = list(product(range(chasm_left_x + chasm_size, room.width), range(room.length)))
        # Target range excludes positions next to walls to avoid grabbing food through walls
        target_x_range = range(1, chasm_left_x - 1)
        target_region_tiles = list(product(target_x_range, range(1, room.length - 1)))
        spawn_room, spawn_tile = decide_spawn_room_and_tile(rand, first_story, {room.id: spawn_region_tiles})

        target_tile = only(decide_tiles_by_distance(target_region_tiles, spawn_tile, difficulty, rand))
        free_spawn_region_tiles = [
            tile
            for tile in spawn_region_tiles
            if not squares_overlap(
                tile_centroid(spawn_tile), tile_size, tile_centroid(tile, boulder_size), boulder_size
            )
            and (boulder_size <= tile_size or (tile[0] not in {0, room.width - 1}))
        ]
        boulder_tile = only(decide_tiles_by_distance(free_spawn_region_tiles, target_tile, difficulty, rand))

        extra_build_params = dict(
            chasm_left_x=chasm_left_x,
            chasm_size=chasm_size,
            chasm_depth=chasm_depth,
            spawn_tile=spawn_tile,
            boulder_tile=boulder_tile,
            boulder_size=boulder_size,
            target_tile=target_tile,
        )
        return extra_build_params

    def _add_obstacles_blocked_entrance(
        self,
        building: Building,
        room: Room,
        *,
        hole_tile: Tuple[int, int],
        spawn_tile: Tuple[int, int],
        boulder_size: float,
        target_tile: Tuple[int, int],
    ):
        tile_size = 1
        hole_tile_x, hole_tile_z = hole_tile

        wall_point_set = (
            (BuildingTile(x=hole_tile_x, z=0), BuildingTile(x=hole_tile_x, z=hole_tile_z - 1)),
            (BuildingTile(x=hole_tile_x, z=hole_tile_z + 1), BuildingTile(x=hole_tile_x, z=room.length - 1)),
        )

        # Make wall unjumpable, but low enough to potentially see food on other side
        wall_height = room.outer_height
        new_heightmap = room.floor_heightmap.copy()
        windows = []
        for points in wall_point_set:
            wall = Wall(0, 0, points, 1, wall_height)
            draw_line_in_grid(wall.points, new_heightmap, wall.height, drawable_grid_value=None)

            window_y = DEFAULT_FLOOR_THICKNESS + (wall_height - DEFAULT_FLOOR_THICKNESS) / 2
            window_position_room_coords = midpoint(
                np.array([points[0].x + tile_size / 2, window_y, points[0].z]),
                np.array([points[1].x + tile_size / 2, window_y, points[1].z + tile_size]),
            )
            window_length = round((euclidean_distance(points[0], points[1]) + tile_size) * 0.8)
            window_size = wall.thickness, wall_height * 0.5, window_length
            room_offset = np.array([room.position.x, 0, room.position.z])
            window_position = local_to_global_coords(window_position_room_coords, room_offset)
            window = Window(window_position, np.array(window_size))
            windows.append(window)

        initial_room_id = 0
        building.stories[0].rooms[initial_room_id] = room.with_heightmap(new_heightmap)
        building.stories[0] = attr.evolve(building.stories[0], windows=building.stories[0].windows + windows)

        boulder_location_in_room = (
            hole_tile_x + 1 + boulder_size / 2,
            room.floor_heightmap[hole_tile_z, hole_tile_x] + boulder_size / 2 - DEFAULT_FLOOR_THICKNESS,
            hole_tile_z + tile_size / 2,
        )

        spawn_height = room.floor_heightmap[spawn_tile[1], spawn_tile[0]] + AGENT_HEIGHT / 2
        spawn_location_in_room = spawn_tile[0] + tile_size / 2, spawn_height, spawn_tile[1] + tile_size / 2
        target_height = room.floor_heightmap[target_tile[1], target_tile[0]] + FOOD_HOVER_DIST
        target_location_in_room = target_tile[0] + tile_size / 2, target_height, target_tile[1] + tile_size / 2
        return spawn_location_in_room, target_location_in_room, boulder_location_in_room, boulder_size

    def _add_obstacles_cross_chasm(
        self,
        building: Building,
        room: Room,
        *,
        chasm_left_x: int,
        chasm_size: int,
        chasm_depth: float,
        spawn_tile: Tuple[int, int],
        boulder_tile: Tuple[int, int],
        boulder_size: float,
        target_tile: Tuple[int, int],
    ):
        tile_size = 1
        new_heightmap = room.floor_heightmap.copy()
        new_heightmap[0 : room.length, chasm_left_x : chasm_left_x + chasm_size] = -chasm_depth
        room = room.with_heightmap(new_heightmap)
        building.stories[0].rooms[0] = room

        spawn_height = room.floor_heightmap[spawn_tile[1], spawn_tile[0]] + AGENT_HEIGHT / 2
        spawn_location_in_room = spawn_tile[0] + tile_size / 2, spawn_height, spawn_tile[1] + tile_size / 2

        # Prevent reaching by jump+grab
        room, _ = add_food_island(room, BuildingTile(*target_tile), 1, AGENT_HEIGHT)
        building.stories[0].rooms[0] = room
        target_height = room.floor_heightmap[target_tile[1], target_tile[0]] + FOOD_HOVER_DIST
        target_location_in_room = target_tile[0] + tile_size / 2, target_height, target_tile[1] + tile_size / 2

        boulder_height = (
            room.floor_heightmap[boulder_tile[1], boulder_tile[0]] + boulder_size / 2 - DEFAULT_FLOOR_THICKNESS
        )
        boulder_location_in_room = boulder_tile[0] + tile_size / 2, boulder_height, boulder_tile[1] + tile_size / 2
        return spawn_location_in_room, target_location_in_room, boulder_location_in_room, boulder_size

    def _add_obstacles_climb_higher(
        self,
        building: Building,
        room: Room,
        *,
        spawn_tile: Tuple[int, int],
        food_island_top_left_tile: Tuple[int, int],
        food_island_size: int,
        food_island_height: float,
        boulder_tile: Tuple[int, int],
        boulder_size: float,
    ):
        tile_size = 1
        min_room_size = math.ceil(boulder_size) * 2 + food_island_size
        if room.width < min_room_size or room.length < min_room_size:
            raise ImpossibleWorldError("Room too small to fit a food island and boulder")

        food_island_position = BuildingTile(*food_island_top_left_tile)
        room_with_island, food_island_tiles = add_food_island(
            room, food_island_position, food_island_size, food_island_height
        )
        room_with_island.outer_height = (
            room_with_island.floor_heightmap.max() + CEILING_THICKNESS + 2 * MAX_JUMP_HEIGHT_METERS
        )
        building.stories[0].rooms[0] = room_with_island

        spawn_location_in_room = (
            spawn_tile[0] + tile_size / 2,
            room_with_island.floor_heightmap[spawn_tile[1], spawn_tile[0]] + AGENT_HEIGHT / 2,
            spawn_tile[1] + tile_size / 2,
        )

        boulder_location_in_room = (
            boulder_tile[0] + tile_size / 2,
            room_with_island.floor_heightmap[boulder_tile[1], boulder_tile[0]]
            - DEFAULT_FLOOR_THICKNESS
            + boulder_size / 2,
            boulder_tile[1] + tile_size / 2,
        )

        food_island_center_tile = (
            food_island_position.x + food_island_size // 2,
            food_island_position.z + food_island_size // 2,
        )
        target_location_in_room = (
            food_island_center_tile[0] + tile_size / 2,
            room_with_island.floor_heightmap[food_island_center_tile[1], food_island_center_tile[0]] + FOOD_HOVER_DIST,
            food_island_center_tile[1] + tile_size / 2,
        )

        return spawn_location_in_room, target_location_in_room, boulder_location_in_room, boulder_size

    def add_principal_obstacles(
        self, rand: np.random.Generator, building: Building, obstacle_params: Tuple
    ) -> IndoorTaskParams:
        build_fn, extra_build_params, boulder_mass, entrance_azimuth = obstacle_params

        spawn_story = building.stories[0]
        spawn_room = spawn_story.rooms[0]
        spawn_location_in_room, target_location_in_room, boulder_location_in_room, boulder_size = build_fn(
            building, spawn_room, **extra_build_params
        )

        room_offset = np.array([spawn_room.position.x, 0, spawn_room.position.z])
        spawn_location_in_building = local_to_global_coords(np.array(spawn_location_in_room), room_offset)
        target_location_in_building = local_to_global_coords(np.array(target_location_in_room), room_offset)
        boulder_location_in_building = local_to_global_coords(np.array(boulder_location_in_room), room_offset)

        entities = [Boulder(position=boulder_location_in_building, mass=boulder_mass, size=boulder_size)]
        entrance_sites = ((spawn_story.num, spawn_room.id, tuple([Azimuth.EAST])),)
        return building, entities, spawn_location_in_building, target_location_in_building, entrance_sites



import os
import time
import traceback
from multiprocessing import Pool
from pathlib import Path
from typing import Dict
from typing import Final
from typing import Generic
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import TypeVar

import attr
import numpy as np
from loguru import logger

from avalon.common.log_utils import configure_parent_logging
from avalon.common.log_utils import logger
from avalon.datagen.errors import ImpossibleWorldError
from avalon.datagen.world_creation.configs.export import get_eval_agent_export_config
from avalon.datagen.world_creation.configs.export import get_oculus_export_config
from avalon.datagen.world_creation.constants import AvalonTask
from avalon.datagen.world_creation.entities.animals import ALL_PREDATOR_CLASSES
from avalon.datagen.world_creation.entities.animals import ALL_PREY_CLASSES
from avalon.datagen.world_creation.entities.food import FOODS
from avalon.datagen.world_creation.tasks.eat import ForcedFood
from avalon.datagen.world_creation.tasks.fight import ForceFight
from avalon.datagen.world_creation.tasks.hunt import ForceHunt
from avalon.datagen.world_creation.types import DebugVisualizationConfig
from avalon.datagen.world_creation.world_generator import GENERATION_FUNCTION_BY_TASK
from avalon.datagen.world_creation.world_generator import GenerateAvalonWorldParams
from avalon.datagen.world_creation.worlds.export import get_world_slug

ForcedParams = TypeVar("ForcedParams")


@attr.s(auto_attribs=True, collect_by_mro=True)
class _ForceGeneratedWorldCompleteness(Generic[ForcedParams]):
    """Ensures generated worlds include certain forced configurations to guarantee completeness."""

    forced_params: List[ForcedParams]
    difficulty: float

    @property
    def difficulties(self) -> List[float]:
        return [self.difficulty] * len(self.forced_params)


_DIFFICULT_WORLD_FRACTION = 0.2


_SPECIAL_COMPLETENESS_CONFIGS: Final[Dict[AvalonTask, _ForceGeneratedWorldCompleteness]] = {
    AvalonTask.FIGHT: _ForceGeneratedWorldCompleteness(
        difficulty=1.0,
        forced_params=[
            ForceFight(
                weapon_value=2,
                large_weapon_probability=0,
                rock_probability=1.0,
                predators=(predator, predator, predator),
            )
            for predator in ALL_PREDATOR_CLASSES
        ],
    ),
    AvalonTask.HUNT: _ForceGeneratedWorldCompleteness(
        difficulty=1.0,
        forced_params=[
            ForceHunt(
                weapon_value=2,
                large_weapon_probability=0,
                rock_probability=1.0,
                prey=prey,
            )
            for prey in ALL_PREY_CLASSES
        ],
    ),
    AvalonTask.EAT: _ForceGeneratedWorldCompleteness(
        difficulty=0.5,
        forced_params=[ForcedFood(food=x.__class__) for x in FOODS],
    ),
}


def _resolve_forced_difficulties_for_completeness(task: AvalonTask, is_practice: bool) -> List[float]:
    if is_practice or task not in _SPECIAL_COMPLETENESS_CONFIGS:
        return []
    return _SPECIAL_COMPLETENESS_CONFIGS[task].difficulties


def _resolve_forced_kwargs_for_completeness(task: AvalonTask, is_practice: bool, task_index: int) -> dict:
    if is_practice or task not in _SPECIAL_COMPLETENESS_CONFIGS:
        return {}

    forced_params_for_task = _SPECIAL_COMPLETENESS_CONFIGS[task].forced_params
    if task_index < len(forced_params_for_task):
        return {"_FORCED": forced_params_for_task[task_index]}

    return {}


def get_difficulties_for_task(task: AvalonTask, is_practice: bool, min_difficulty: float, num_worlds_per_task: int):
    forced_difficulties_for_completeness = _resolve_forced_difficulties_for_completeness(task, is_practice)
    remaining_world_count = num_worlds_per_task - len(forced_difficulties_for_completeness)
    if remaining_world_count <= 0:
        return forced_difficulties_for_completeness[:num_worlds_per_task]
    difficult_world_count = round(remaining_world_count * _DIFFICULT_WORLD_FRACTION)
    remaining_world_count -= difficult_world_count
    return (
        forced_difficulties_for_completeness
        + [round(x.item(), 2) for x in np.linspace(min_difficulty, 1.0, remaining_world_count)]
        + difficult_world_count * [1.0]
    )


class GeneratedWorld(NamedTuple):
    world_id: str
    task: AvalonTask
    difficulty: float
    time: float
    impossible_world_errors: List[Tuple[str, str]]
    unhandled_errors: List[Tuple[str, str]]
    was_successful: bool


def _generate_world(
    base_output_path: Path,
    task: AvalonTask,
    difficulty: float,
    seed: int,
    world_id: str,
    # only used by special task generators so that they can know when they're special
    task_idx: int,
    is_practice: bool,
    debug_visualization_config: Optional[DebugVisualizationConfig],
    is_generating_for_human: bool,
    max_retries: int = 10,
) -> GeneratedWorld:
    if is_generating_for_human:
        export_config = get_oculus_export_config(world_id)
    else:
        export_config = get_eval_agent_export_config()
    if debug_visualization_config:
        export_config = attr.evolve(export_config, is_tiled=False)
        export_config = attr.evolve(export_config, debug_visualization_config=debug_visualization_config)
    start_time = time.time()
    rand = np.random.default_rng(seed)

    output_path = base_output_path / world_id
    output_path.mkdir(parents=True, exist_ok=True)
    generation_function = GENERATION_FUNCTION_BY_TASK[task]

    impossible_world_errors = []
    unhandled_errors = []

    was_successful = True

    kwargs = _resolve_forced_kwargs_for_completeness(task, is_practice, task_idx)

    i = 0
    while i < max_retries:
        try:
            generation_function(rand, difficulty, output_path, export_config, **kwargs)
            break
        except ImpossibleWorldError as e:
            impossible_world_errors.append((world_id, str(e)))
        except Exception as e:
            unhandled_errors.append((world_id, str(e)))
            traceback.print_exc()
            raise
        i += 1

    if i >= max_retries:
        was_successful = False

    end_time = time.time()
    total_time = end_time - start_time
    return GeneratedWorld(
        world_id=world_id,
        task=task,
        difficulty=difficulty,
        time=total_time,
        impossible_world_errors=impossible_world_errors,
        unhandled_errors=unhandled_errors,
        was_successful=was_successful,
    )


def generate_evaluation_worlds(
    base_output_path: Path,
    tasks: Sequence[AvalonTask],
    num_worlds_per_task: int,
    is_generating_for_human: bool,
    start_seed: int = 0,
    is_practice: bool = False,
    min_difficulty: float = 0.0,
    is_recreating: bool = True,
    num_workers: int = 10,
    debug_visualization_config: Optional[DebugVisualizationConfig] = None,
    is_async: bool = True,
    is_verbose: bool = True,
) -> List[GeneratedWorld]:
    """Generates worlds for evaluation and testing, either by models or humans.

    Ensures a special set of configurations are applied to guarantee a completeness of variety in certain tasks (disabled by is_practice).
    For instance, the generated set will include one of each food in EAT, a difficult FIGHT with each predator, etc.
    """
    assert (
        os.getenv("PYTHONHASHSEED", None) is not None
    ), f"PYTHONHASHSEED must be set for worlds to be generated deterministically."
    total_errors = 0
    all_impossible_world_errors = []
    all_unhandled_errors = []
    total_error_counter = [0]

    def on_done(result: GeneratedWorld) -> None:
        if is_verbose:
            logger.info(f"Finished generating {result.world_id}")
        if not result.was_successful:
            total_error_counter[0] = total_error_counter[0] + 1
        all_impossible_world_errors.extend(result.impossible_world_errors)
        all_unhandled_errors.extend(result.unhandled_errors)

    errors = []

    def on_error(error: BaseException) -> None:
        errors.append(error)
        logger.error("BAD: one of the level generators failed")

    start_time = time.time()
    results = []

    params_to_generate: List[GenerateAvalonWorldParams] = []
    current_seed = start_seed
    for task in tasks:
        difficulties = get_difficulties_for_task(task, is_practice, min_difficulty, num_worlds_per_task)
        task_name = task.value.lower()
        for i in range(num_worlds_per_task):
            difficulty = difficulties[i]
            world_id = get_world_slug(task_name, current_seed, difficulty, is_practice)
            params_to_generate.append(
                GenerateAvalonWorldParams(
                    seed=current_seed,
                    index=i,
                    difficulty=difficulty,
                    output=str(base_output_path / world_id),
                    task=task,
                )
            )
            current_seed += 1

    with Pool(processes=num_workers, initializer=configure_parent_logging) as worker_pool:
        requests = []
        for params in params_to_generate:
            world_id = params.output_path.name
            if not is_recreating and params.output_path.exists():
                continue

            args = (
                base_output_path,
                params.task,
                params.difficulty,
                params.seed,
                world_id,
                params.index,
                is_practice,
                debug_visualization_config,
                is_generating_for_human,
            )

            if is_async:
                request = worker_pool.apply_async(
                    _generate_world,
                    args=args,
                    callback=on_done,
                    error_callback=on_error,
                )
                requests.append(request)
            else:
                results.append(_generate_world(*args))

        for request in requests:
            request.wait()
            if request._success:  # type: ignore[attr-defined]
                results.append(request.get())
    total_num_worlds = num_worlds_per_task * len(tasks)

    if len(errors) > 0:
        raise errors[0]

    logger.info(
        f"Finished generating {total_num_worlds - total_errors} of {total_num_worlds} worlds in {(time.time() - start_time) / 60:.2f} minutes"
    )
    for err in all_impossible_world_errors:
        logger.info(f"Failed with impossible world error:\n{err}")

    for err in all_unhandled_errors:
        logger.warning(f"Failed with unhandled error:\n{err}")

    return results



import glob
import hashlib
import itertools
import os
import time
from enum import Enum
from hashlib import md5
from itertools import groupby
from pathlib import Path
from typing import Any
from typing import Callable
from typing import Dict
from typing import Final
from typing import Iterable
from typing import Iterator
from typing import KeysView
from typing import List
from typing import Optional
from typing import Protocol
from typing import Tuple
from typing import TypeVar
from typing import Union
from typing import ValuesView
from typing import cast

import numpy as np
from IPython.display import display

from avalon.contrib.utils import FILESYSTEM_ROOT
from avalon.contrib.utils import TEMP_DIR

TMP_DATA_DIR: Final = os.path.join(TEMP_DIR, "wandb")

AVALON_PACKAGE_DIR: Final = os.path.abspath(f"{__file__}/../..")


class _SupportsLessThan(Protocol):
    def __lt__(self, __other: Any) -> bool:
        ...


T = TypeVar("T")
TK = TypeVar("TK", bound=_SupportsLessThan)
TV = TypeVar("TV")


class EvaluationStage(Enum):
    TRAINING = "TRAIN"
    VALIDATION = "VAL"
    TEST = "TEST"
    GENERATION = "GEN"


def show(x: Any, **kwargs):  # type: ignore
    """This wrapper is just here so it shuts up about the type signature, it can display anything"""
    return display(x, **kwargs)


def create_random_image() -> None:
    import numpy
    from PIL import Image

    image_array = numpy.random.rand(512, 512, 3) * 255
    image = Image.fromarray(image_array.astype("uint8")).convert("RGBA")
    image.save("random_512.png")


def partition_dict(pred_fn: Callable, d: Dict) -> Tuple[Dict, Dict]:
    """
    Use a predicate to partition a dictionary into two mutually exclusive sub-dicts with values that tested False, True
    partition_dict(is_odd, {'a': 0, 'b': 1}) --> ({'a': 0}, {'b': 1})
    """
    return {k: v for k, v in d.items() if not pred_fn(k, v)}, {k: v for k, v in d.items() if pred_fn(k, v)}


def flatten(iterable: Iterable[Iterable[T]]) -> List[T]:
    return list(itertools.chain(*iterable))


def compact(x: Any) -> Any:
    return list(filter(None, x))


def group_by_helper(data: Iterable[TV], get_key: Callable[[TV], TK]) -> Dict[TK, List[TV]]:
    data = sorted(data, key=get_key)
    return {k: list(g) for k, g in groupby(data, get_key)}


def first(iterator: Union[Iterable[T], Iterator[T]]) -> Optional[T]:
    if isinstance(iterator, (tuple, list, KeysView, ValuesView)):
        return next(iter(iterator), None)
    elif hasattr(iterator, "__next__"):
        return next(cast(Iterator[T], iterator), None)
    elif isinstance(iterator, np.ndarray):
        if len(iterator) == 0:
            return None
        return cast(T, iterator[0])
    elif iterator is None:
        raise Exception("Not a sequence, is None!")
    raise Exception("Ooops, handle this")


def only(x: Iterable[T]) -> T:
    materialized_x = list(x)
    assert len(materialized_x) == 1
    return materialized_x[0]


def get_root_experiment_dir() -> Path:
    return Path(os.getcwd().split("/current")[0])


def get_path_to_current_dir() -> Path:
    return get_root_experiment_dir() / "current"


def get_relative_project_path() -> str:
    project_dir = find_root_dir()
    git_dir = find_git_dir()
    return str(Path(project_dir).relative_to(Path(git_dir)))


def find_root_dir() -> str:
    return find_dir_by_file_name("root_dir_marker.txt")


def find_git_dir() -> str:
    return find_dir_by_file_name(".gitignore")


def find_dir_by_file_name(target_file_name: str) -> str:
    filepath_parts = str(Path(__file__).absolute()).split("/")
    for dirname in filepath_parts[1:]:
        partial_dirpath = filepath_parts[: filepath_parts.index(dirname)] + [target_file_name]
        if os.path.exists("/".join(partial_dirpath)):
            return os.path.dirname("/".join(partial_dirpath))
    raise Exception(f"Can't find {target_file_name}")


def hash_dirpath(dirpath: str) -> str:
    files = sorted(glob.glob(dirpath, recursive=True))
    return md5(
        "".join([md5(open(x, "rb").read()).hexdigest() for x in files if os.path.isfile(x)]).encode("utf-8")
    ).hexdigest()


def conv_transpose_out_size(
    in_size: int, stride: int, padding: int, kernel_size: int, out_padding: int, dilation: int = 1
) -> int:
    return (in_size - 1) * stride - 2 * padding + dilation * (kernel_size - 1) + out_padding + 1


def conv_out_size(in_size: int, stride: int, padding: int, kernel_size: int) -> int:
    # note: this is missing dilation but we don't normally change it
    return int((in_size - kernel_size + 2 * padding) / stride + 1.0)


DATA_FOLDER = os.path.join(FILESYSTEM_ROOT, "data")


FIX_ENCODER_STRINGS = {"0": "conv", "1": "bn"}


def _fix_key(key: str, prefix: str) -> str:
    parts_without_prefix = key[len(prefix) + 1 :].split(".")
    if prefix == "encoder" and len(parts_without_prefix) > 2:
        layer_part = parts_without_prefix[2]
        if layer_part in FIX_ENCODER_STRINGS:
            parts_without_prefix[2] = FIX_ENCODER_STRINGS[layer_part]
    return ".".join(parts_without_prefix)


def float_to_str(f: float) -> str:
    # rstrip is needed to account for int cases
    return np.format_float_positional(f).rstrip(".").replace(".", "_")


def to_immutable_array(value: np.ndarray) -> np.ndarray:
    value.setflags(write=False)
    return value


def file_checksum(file_path: str) -> str:
    with open(file_path, "rb") as output_file:
        data = output_file.read()
        return hashlib.md5(data).hexdigest()


def dir_checksum(dir_path: Path, glob: str = "*") -> str:
    hash_md5 = hashlib.md5()
    for path in sorted(dir_path.glob(glob)):
        with open(path, "rb") as f:
            for chunk in iter(lambda: f.read(2**20), b""):
                hash_md5.update(chunk)
    return hash_md5.hexdigest()


def wait_until_true(
    callback: Callable[[], Optional[bool]],
    max_wait_sec: float = 5,
    sleep_inc: float = 0.25,
):
    """Repeatedly call callback() until it returns True or max_wait_sec is reached"""
    waited_for_sec = 0.0000
    while waited_for_sec <= max_wait_sec:
        if callback():
            return
        time.sleep(sleep_inc)
        waited_for_sec += sleep_inc
    raise TimeoutError(f"could not complete within {max_wait_sec} seconds")



from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import TypeVar

import numpy as np


def get_difficulty_based_value(
    difficulty: float, min_val: float, max_val: float, variability: float, rand: np.random.Generator
) -> float:
    total_delta = max_val - min_val
    delta = variability * total_delta
    remainder = total_delta - delta
    return min_val + (remainder * difficulty) + (rand.uniform() * delta)


T = TypeVar("T")


def select_categorical_difficulty(
    choices: Sequence[T],
    difficulty: float,
    rand: np.random.Generator,
    _FORCED: Optional[T] = None,
) -> Tuple[T, float]:
    """
    returns selected choice, new difficulty
    """

    num_choices = len(choices)
    prob_coeff = 1.0 / sum(difficulty ** float(x) for x in range(num_choices))
    choice_prob = [
        difficulty / num_choices + (1 - difficulty) * (difficulty**x) * prob_coeff for x in range(num_choices)
    ]
    choice_idx = rand.choice(range(num_choices), p=choice_prob)
    if _FORCED is not None:
        assert _FORCED in choices
        choice_idx = choices.index(_FORCED)

    # TODO use other less arbitrary method to calculate?
    new_difficulty = difficulty**2 + (1 - difficulty) * (difficulty ** (choice_idx + 1))
    return choices[choice_idx], new_difficulty


def select_boolean_difficulty(
    difficulty: float,
    rand: np.random.Generator,
    initial_prob: float = 1,
    final_prob: float = 0.01,
    _FORCED: Optional[bool] = None,
) -> Tuple[bool, float]:
    """
    Uses log interpolation to scale initial prob (difficulty=0) to final prob (difficulty=1). Zero probability is not
    valid, but 1 is. Rescales difficulty to account for probability distribution.
    Tips:
    - Always finish boolean choices before using scalars
    - Somewhat better to have easier option be "true", since we can set p(True) = 1 at difficulty 0.

    debug_override_value that is not None will ALWAYS return that value!
    """
    sampled_value = rand.uniform()
    if _FORCED is not None:
        sampled_value = 0.0 if _FORCED else 1.0

    assert initial_prob <= 1 and final_prob <= 1, "Probability cannot be greater than 1!"
    if initial_prob == final_prob:
        return sampled_value < initial_prob, difficulty
    assert initial_prob > 0 and final_prob > 0, "Cannot have zero probability for True with log interpolation!"
    prob = (initial_prob ** (1 - difficulty)) * (final_prob**difficulty)
    value = sampled_value < prob
    # the updated difficulty is the integral of the probability to d over the integral to 1
    if value:
        new_difficulty = (prob - initial_prob) / (final_prob - initial_prob)
    else:
        prob_coeff = np.log(initial_prob) - np.log(final_prob)
        new_difficulty = (prob - initial_prob + difficulty * prob_coeff) / (final_prob - initial_prob + prob_coeff)
    # the abs here ensures that we don't return -0.0, which caused an annoying bug before
    return value, np.clip(new_difficulty, 0.0, 1.0)


def scale_with_difficulty(
    difficulty: float, start_val: float, end_val: float, _FORCED: Optional[float] = None
) -> float:
    if _FORCED:
        return _FORCED
    delta = end_val - start_val
    return start_val + delta * difficulty


def difficulty_variation(
    start_val: float,
    end_val: float,
    rand: np.random.Generator,
    difficulty: float,
    _FORCED: Optional[float] = None,
) -> float:
    if _FORCED:
        return _FORCED
    delta = end_val - start_val
    return start_val + delta * difficulty * rand.uniform()


def normal_distrib_range(
    start_val: float,
    end_val: float,
    std_dev: float,
    rand: np.random.Generator,
    difficulty: float,
    _FORCED: Optional[float] = None,
) -> float:
    if _FORCED:
        return _FORCED
    delta = end_val - start_val
    mean = start_val + delta * difficulty
    min_val = min([start_val, end_val])
    max_val = max([start_val, end_val])
    return float(np.clip(rand.normal(mean, std_dev), min_val, max_val))


def get_rock_probability(difficulty: float) -> float:
    return scale_with_difficulty(difficulty, 0.5, 0.95)



from typing import Any
from typing import Callable
from typing import Dict
from typing import List

import numpy as np
import torch
import tree
from torch import Tensor
from torch.nn import functional as F
from tree import map_structure

# "TanhBijector" and "SampleDist" are from the following repo
# https://github.com/juliusfrost/dreamer-pytorch


class SampleDist(torch.distributions.Distribution):
    def __init__(self, dist: torch.distributions.Distribution, samples: int = 100) -> None:
        self._dist = dist
        self._samples = samples

    @property
    def name(self) -> str:
        return "SampleDist"

    @property
    def mean(self) -> Tensor:
        sample = self._dist.rsample()
        return torch.mean(sample, 0)

    def mode(self) -> Tensor:
        dist = self._dist.expand((self._samples, *self._dist.batch_shape))
        sample = dist.rsample()
        logprob = dist.log_prob(sample)
        batch_size = sample.size(1)
        feature_size = sample.size(2)
        indices = torch.argmax(logprob, dim=0).reshape(1, batch_size, 1).expand(1, batch_size, feature_size)
        return torch.gather(sample, 0, indices).squeeze(0)

    def entropy(self) -> Tensor:
        dist = self._dist.expand((self._samples, *self._dist.batch_shape))
        sample = dist.rsample()
        logprob = dist.log_prob(sample)
        return -torch.mean(logprob, 0)

    def log_prob(self, sample: Tensor) -> Tensor:
        return self._dist.log_prob(sample)  # type: ignore

    def sample(self) -> Tensor:  # type: ignore
        return self._dist.sample().detach()  # type: ignore

    def rsample(self) -> Tensor:  # type: ignore
        return self._dist.rsample()  # type: ignore


class TanhBijector(torch.distributions.Transform):
    def __init__(self) -> None:
        super().__init__()
        self.bijective = True
        self.domain = torch.distributions.constraints.real
        self.codomain = torch.distributions.constraints.interval(-1.0, 1.0)

    @property
    def sign(self) -> float:
        return 1.0

    def _call(self, x: Tensor) -> Tensor:
        return torch.tanh(x)

    def _inverse(self, y: Tensor) -> Tensor:
        # TODO: cast this to fp32 if we do half-precision
        y = torch.where((torch.abs(y) <= 1.0), torch.clamp(y, -0.99999997, 0.99999997), y)
        y = torch.atanh(y)
        return y

    def log_abs_det_jacobian(self, x: Tensor, y: Tensor) -> Tensor:
        return 2.0 * (np.log(2) - x - F.softplus(-2.0 * x))  # type: ignore


def lambda_return(
    reward: Tensor, value: Tensor, pcont: Tensor, bootstrap: Tensor, lambda_: float, axis: int
) -> Tensor:
    """This is dreamer's value backup method. Similar to GAE - is it the same?
    # Setting lambda=1 gives a discounted Monte Carlo return.
    # Setting lambda=0 gives a fixed 1-step return.
    """
    assert axis == 0
    assert isinstance(pcont, torch.Tensor)
    assert len(reward.shape) == len(value.shape), (reward.shape, value.shape)
    assert bootstrap is not None
    # tensor[None] just adds a dim of len 1 at the beginning of the tensor. so x, y, z => 1, x, y z
    next_values = torch.cat([value[1:], bootstrap[None]], 0)
    inputs = reward + pcont * next_values * (1 - lambda_)
    # Here the map structures are: tuple of tensors, tensor.
    returns: Tensor = static_scan(
        lambda agg, cur: cur[0] + cur[1] * lambda_ * agg, (inputs, pcont), bootstrap, reverse=True
    )
    return returns


def static_scan(fn: Callable, inputs: Any, start: Any, reverse: bool = False) -> Any:
    """This is a utility from danijar for applying a function iteratively over each element in a sequence,
    combined with the current value of an accumulator. It's quite confusing but also quite handy."""
    last = start
    outputs: List[List] = [[] for _ in tree.flatten(start)]
    indices = list(range(len(tree.flatten(inputs)[0])))
    if reverse:
        indices = list(reversed(indices))
    for index in indices:
        inp = map_structure(lambda x: x[index], inputs)
        last = fn(last, inp)
        for o, l in zip(outputs, tree.flatten(last)):
            o.append(l)
    if reverse:
        outputs = [list(reversed(x)) for x in outputs]
    outputs_stacked = [torch.stack(x, 0) for x in outputs]
    return tree.unflatten_as(start, outputs_stacked)


def pack_list_of_dicts(data: List[Dict[str, Tensor]]) -> Dict[str, Tensor]:
    out = {}
    for key in data[0].keys():
        out[key] = torch.stack([x[key] for x in data], dim=0)
    return out



# GENERATED FILE
# See generate_godot_code.py for details

from typing import Final
from typing import Optional

import attr

from avalon.datagen.godot_base_types import *

BRIDGE_LOG_SIGNAL: Final = "Establishing Bridge"
READY_LOG_SIGNAL: Final = "GODATA!"
FIXED_FPS: Final = 10
RGBD_FEATURE: Final = "rgbd"
ISOMETRIC_RGBD_FEATURE: Final = "isometric_rgbd"
TOP_DOWN_RGBD_FEATURE: Final = "top_down_rgbd"
EPISODE_ID_FEATURE: Final = "episode_id"
FRAME_ID_FEATURE: Final = "frame_id"
EVENT_HAPPENED_FEATURE: Final = "event_happened"
ACTION_FEATURE: Final = "action"
RESET_MESSAGE: Final = 0
SEED_MESSAGE: Final = 1
RENDER_MESSAGE: Final = 2
ACTION_MESSAGE: Final = 3
SELECT_FEATURES_MESSAGE: Final = 4
QUERY_AVAILABLE_FEATURES_MESSAGE: Final = 5
CLOSE_MESSAGE: Final = 6
DEBUG_CAMERA_ACTION_MESSAGE: Final = 7
HUMAN_INPUT_MESSAGE: Final = 9
SAVE_SNAPSHOT_MESSAGE: Final = 10
LOAD_SNAPSHOT_MESSAGE: Final = 11
VR_ACTION_SPACE: Final = "VR_ACTION_SPACE"
MOUSE_KEYBOARD_ACTION_SPACE: Final = "MOUSE_KEYBOARD_ACTION_SPACE"
LEFT_HAND: Final = "Left"
RIGHT_HAND: Final = "Right"
FAKE_TYPE_IMAGE: Final = -1
SNAPSHOT_JSON: Final = "snapshot_context.json"
SNAPSHOT_SUBPATH: Final = "snapshots"
SCENE_ROOT_NODE_PATH: Final = "/root/scene_root"
WORLD_NODE_PATH: Final = SCENE_ROOT_NODE_PATH + "/world"
DYNAMIC_TRACKER_NODE_NAME: Final = "dynamic_tracker"
DYNAMIC_TRACKER_NODE_PATH: Final = WORLD_NODE_PATH + "/Avalon/dynamic_tracker"


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class SpecBase(Serializable):
    pass


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class ControlledNodeSpec(SpecBase):
    spawn_point_name: str


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class RecordingOptionsSpec(SpecBase):
    user_id: Optional[str]
    apk_version: Optional[str]
    recorder_host: Optional[str]
    recorder_port: Optional[int]
    is_teleporter_enabled: bool
    resolution_x: int
    resolution_y: int
    is_recording_human_actions: bool
    is_remote_recording_enabled: bool
    is_adding_debugging_views: bool
    is_debugging_output_requested: bool


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class PlayerSpec(ControlledNodeSpec):
    # m/s
    max_head_linear_speed: float
    # deg/s
    max_head_angular_speed: AnyVec3
    # m/s
    max_hand_linear_speed: float
    # deg/s
    max_hand_angular_speed: float
    # m
    jump_height: float
    arm_length: float
    starting_hit_points: float
    # kg
    mass: float
    arm_mass_ratio: float
    head_mass_ratio: float
    # m/s
    standup_speed_after_climbing: float
    # m
    min_head_position_off_of_floor: float
    # N
    push_force_magnitude: float
    # N
    throw_force_magnitude: float
    starting_left_hand_position_relative_to_head: AnyVec3
    starting_right_hand_position_relative_to_head: AnyVec3
    # m/s
    minimum_fall_speed: float
    fall_damage_coefficient: float
    num_frames_alive_after_food_is_gone: int
    eat_area_radius: float
    is_displaying_debug_meshes: bool
    is_human_playback_enabled: bool
    is_slowed_from_crouching: bool


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class AgentPlayerSpec(PlayerSpec):
    pass


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class HumanPlayerSpec(PlayerSpec):
    pass


# TODO: should SimSpec still inherit from DataConfigImplementation?
@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class SimSpec(SpecBase, DataConfigImplementation):
    episode_seed: Optional[int]
    dir_root: str
    recording_options: RecordingOptionsSpec
    player: PlayerSpec


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class AvalonSimSpec(SimSpec):
    pass


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class MouseKeyboardAgentPlayerSpec(AgentPlayerSpec):
    pass


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class MouseKeyboardHumanPlayerSpec(HumanPlayerSpec):
    pass


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class VRAgentPlayerSpec(AgentPlayerSpec):
    pass


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class VRHumanPlayerSpec(HumanPlayerSpec):
    pass


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class CaregiverSimSpec(AvalonSimSpec):
    caregiver: ControlledNodeSpec
